import scipy
import scipy.sparse
import numpy as np
import copy


def util_convert_csr_to_dds(adj_scipy_csr, num_col_partitions):
    """Doing src vertex partitioning (column dimension) by converting csr to dds (dense-dense-sparse).

    Parameters
    ----------
    adj_scipy_csr : scipy.sparse.csr_matrix
        The input matrix to be partitioned

    num_col_partitions : int
        Number of partitions along the column dimension

    Returns
    -------
    s1_pos : numpy.array, dtype is int32
        1-D with shape [num_col_partitions * (num_rows + 1)]

    s1_idx : numpy.array, dtype is int32
        1-D with shape [nnz]

    vals : numpy.array, dtype is float32
        1-D with shape [nnz]
    """
    num_rows = adj_scipy_csr.shape[0]
    num_cols = adj_scipy_csr.shape[1]
    adj_data = adj_scipy_csr.data
    adj_indices = adj_scipy_csr.indices
    adj_indptr = adj_scipy_csr.indptr
    d1_size = num_col_partitions
    d2_size = adj_indptr.shape[0]

    s1_pos = np.zeros(shape=(d1_size*d2_size), dtype=adj_indptr.dtype)
    s1_idx = np.zeros(shape=adj_indices.shape, dtype=adj_indices.dtype)
    vals = np.zeros(shape=adj_data.shape, dtype=adj_data.dtype)

    counter = 0
    num_cols_per_partition = (num_cols + num_col_partitions - 1) // num_col_partitions
    for i in range(num_col_partitions):
        if i == num_col_partitions - 1:
            adj_partition_scipy_csr = adj_scipy_csr[:, (i*num_cols_per_partition)::]
        else:
            adj_partition_scipy_csr = adj_scipy_csr[:, (i*num_cols_per_partition):(i+1)*num_cols_per_partition]
        nnz = adj_partition_scipy_csr.data.shape[0]
        vals[counter:(counter+nnz)] = adj_partition_scipy_csr.data
        s1_pos[i*d2_size:(i+1)*d2_size] = adj_partition_scipy_csr.indptr + counter
        s1_idx[counter:(counter+nnz)] = adj_partition_scipy_csr.indices
        counter += nnz

    assert len(s1_idx) == counter
    assert len(vals) == counter

    return s1_pos, s1_idx, vals


def util_partition_adj_coo_2d(adj_scipy_coo, num_row_partitions, num_col_partitions):
    """Doing 2D partitioning.

    Parameters
    ----------
    adj_scipy_coo : scipy.sparse.coo_matrix
        The input matrix to be partitioned

    num_row_partitions : int
        Number of partitions along the row dimension

    num_col_partitions : int
        Number of partitions along the col dimension

    Returns
    -------
    edge_id_list_after_partition : numpy.array, dtype is int32
        1-D with shape [nnz] (COO)

    adj_row_indices_after_partition : numpy.array, dtype is int32
        1-D with shape [nnz] (COO)

    adj_col_indices_after_partition : numpy.array, dtype is int32
        1-D with shape [nnz] (COO)
    """
    num_rows = adj_scipy_coo.shape[0]
    num_cols = adj_scipy_coo.shape[1]
    adj_row_indices = adj_scipy_coo.row
    adj_col_indices = adj_scipy_coo.col
    nnz = adj_row_indices.shape[0]
    assert adj_col_indices.shape[0] == nnz, "length of adj_row_indices and that of adj_col_indices do not match"
    # Give each edge an id to record the graph traversal order
    adj_scipy_coo.data = np.arange(1, 1 + nnz, dtype='int32')

    # COO matrix is not subscriptable; we need CSR to do partitioning
    adj_scipy_csr = adj_scipy_coo.tocsr()
    edge_id_list_after_partition = np.zeros(shape=(nnz), dtype=adj_scipy_coo.data.dtype)
    adj_row_indices_after_partition = np.zeros(shape=(nnz), dtype=adj_row_indices.dtype)
    adj_col_indices_after_partition = np.zeros(shape=(nnz), dtype=adj_col_indices.dtype)

    num_rows_per_partition = (num_rows + num_row_partitions - 1) // num_row_partitions
    num_cols_per_partition = (num_cols + num_col_partitions - 1) // num_col_partitions
    counter = 0
    for row_idx in range(num_row_partitions):
        for col_idx in range(num_col_partitions):
            if row_idx < num_row_partitions - 1 and col_idx < num_col_partitions - 1:
                adj_partition_scipy_csr = adj_scipy_csr[row_idx*num_rows_per_partition:(row_idx+1)*num_rows_per_partition, \
                    col_idx*num_cols_per_partition:(col_idx+1)*num_cols_per_partition]
            elif row_idx < num_row_partitions - 1 and col_idx == num_col_partitions - 1:
                adj_partition_scipy_csr = adj_scipy_csr[row_idx*num_rows_per_partition:(row_idx+1)*num_rows_per_partition, \
                    col_idx*num_cols_per_partition::]
            elif row_idx == num_row_partitions - 1 and col_idx < num_col_partitions - 1:
                adj_partition_scipy_csr = adj_scipy_csr[row_idx*num_rows_per_partition::, \
                    col_idx*num_cols_per_partition:(col_idx+1)*num_cols_per_partition]
            elif row_idx == num_row_partitions - 1 and col_idx == num_col_partitions - 1:
                adj_partition_scipy_csr = adj_scipy_csr[row_idx*num_rows_per_partition::, \
                    col_idx*num_cols_per_partition::]
            else:
                raise RuntimeError("no condition is satisfied")
            adj_partition_scipy_coo = adj_partition_scipy_csr.tocoo()
            nnz_in_this_partition = adj_partition_scipy_coo.nnz
            edge_id_list_after_partition[counter:(counter+nnz_in_this_partition)] = adj_partition_scipy_coo.data
            adj_row_indices_after_partition[counter:(counter+nnz_in_this_partition)] = \
                adj_partition_scipy_coo.row + row_idx * num_rows_per_partition
            adj_col_indices_after_partition[counter:(counter+nnz_in_this_partition)] = \
                adj_partition_scipy_coo.col + col_idx * num_cols_per_partition
            counter += nnz_in_this_partition

    return edge_id_list_after_partition, adj_row_indices_after_partition, adj_col_indices_after_partition


def util_partition_adj_coo_hilbert(adj_scipy_coo, hilbert_depth):
    num_rows = adj_scipy_coo.shape[0]
    num_cols = adj_scipy_coo.shape[1]
    adj_row_indices = adj_scipy_coo.row
    adj_col_indices = adj_scipy_coo.col
    nnz = adj_row_indices.shape[0]
    assert adj_col_indices.shape[0] == nnz, "length of adj_row_indices and that of adj_col_indices do not match"

    num_row_partitions = pow(2, hilbert_depth)
    num_col_partitions = pow(2, hilbert_depth)

    # Give each edge an id to record the graph traversal order
    adj_scipy_coo.data = np.arange(1, 1 + nnz, dtype='int32')
    # COO matrix is not subscriptable; we need CSR to do partitioning
    adj_scipy_csr = adj_scipy_coo.tocsr()
    edge_id_list_after_partition = np.zeros(shape=(nnz), dtype=adj_scipy_coo.data.dtype)
    adj_row_indices_after_partition = np.zeros(shape=(nnz), dtype=adj_row_indices.dtype)
    adj_col_indices_after_partition = np.zeros(shape=(nnz), dtype=adj_col_indices.dtype)

    hilbert_curve = HilbertCurve(hilbert_depth, 2)

    num_rows_per_partition = (num_rows + num_row_partitions - 1) // num_row_partitions
    num_cols_per_partition = (num_cols + num_col_partitions - 1) // num_col_partitions
    counter = 0
    for partition_idx in range(num_row_partitions * num_col_partitions):
        row_idx, col_idx = hilbert_curve.coordinates_from_distance(partition_idx)
        if row_idx < num_row_partitions - 1 and col_idx < num_col_partitions - 1:
            adj_partition_scipy_csr = adj_scipy_csr[row_idx*num_rows_per_partition:(row_idx+1)*num_rows_per_partition, \
                col_idx*num_cols_per_partition:(col_idx+1)*num_cols_per_partition]
        elif row_idx < num_row_partitions - 1 and col_idx == num_col_partitions - 1:
            adj_partition_scipy_csr = adj_scipy_csr[row_idx*num_rows_per_partition:(row_idx+1)*num_rows_per_partition, \
                col_idx*num_cols_per_partition::]
        elif row_idx == num_row_partitions - 1 and col_idx < num_col_partitions - 1:
            adj_partition_scipy_csr = adj_scipy_csr[row_idx*num_rows_per_partition::, \
                col_idx*num_cols_per_partition:(col_idx+1)*num_cols_per_partition]
        elif row_idx == num_row_partitions - 1 and col_idx == num_col_partitions - 1:
            adj_partition_scipy_csr = adj_scipy_csr[row_idx*num_rows_per_partition::, \
                col_idx*num_cols_per_partition::]
        else:
            raise RuntimeError("no condition is satisfied")
        adj_partition_scipy_coo = adj_partition_scipy_csr.tocoo()
        nnz_in_this_partition = adj_partition_scipy_coo.nnz
        edge_id_list_after_partition[counter:(counter+nnz_in_this_partition)] = adj_partition_scipy_coo.data
        adj_row_indices_after_partition[counter:(counter+nnz_in_this_partition)] = \
            adj_partition_scipy_coo.row + row_idx * num_rows_per_partition
        adj_col_indices_after_partition[counter:(counter+nnz_in_this_partition)] = \
            adj_partition_scipy_coo.col + col_idx * num_cols_per_partition
        counter += nnz_in_this_partition

    return edge_id_list_after_partition, adj_row_indices_after_partition, adj_col_indices_after_partition



import scipy.sparse as sp
import networkx as nx

from cogdl.models import build_model
from cogdl.utils.utils import ArgClass
from cogdl.wrappers.tools.wrapper_utils import evaluate_node_embeddings_using_liblinear
from .. import EmbeddingModelWrapper


class NetworkEmbeddingModelWrapper(EmbeddingModelWrapper):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--num-shuffle", type=int, default=10)
        parser.add_argument("--training-percents", default=[0.9], type=float, nargs="+")
        parser.add_argument("--enhance", type=str, default=None, help="use prone or prone++ to enhance embedding")
        parser.add_argument("--max-evals", type=int, default=10)
        parser.add_argument("--num-workers", type=int, default=1)
        # fmt: on

    def __init__(self, model, num_shuffle=1, training_percents=[0.1], enhance=None, max_evals=10, num_workers=1):
        super(NetworkEmbeddingModelWrapper, self).__init__()
        self.model = model
        self.num_shuffle = num_shuffle
        self.training_percents = training_percents
        self.enhance = enhance
        self.max_evals = max_evals
        self.num_workers = num_workers

    def train_step(self, batch):
        emb = self.model(batch)
        if self.enhance is not None:
            self._enhance_emb(batch, emb)
        return emb

    def test_step(self, batch):
        x, y = batch
        return evaluate_node_embeddings_using_liblinear(x, y, self.num_shuffle, self.training_percents)

    def _enhance_emb(self, graph, embs):
        A = nx.to_scipy_sparse_matrix(graph.to_networkx())
        args = ArgClass()
        if self.enhance == "prone":
            args.model = "prone"
            args.hidden_size = embs.shape[1]
            args.step, args.theta, args.mu = 5, 0.5, 0.2
            model = build_model(args)
            embs = model._chebyshev_gaussian(A, embs)
        elif self.enhance == "prone++":
            args.model = "prone++"
            args.filter_types = ["heat", "ppr", "gaussian", "sc"]
            args.max_evals = self.max_evals
            args.num_workers = self.num_workers
            args.no_svd = False
            args.loss = "infomax"
            args.no_search = False
            model = build_model(args)
            embs = model(embs, A)
        else:
            raise ValueError("only supports 'prone' and 'prone++'")
        return embs



from .base_logger import Logger


def build_logger(logger, log_path="./runs", project="cogdl-exp"):
    if logger == "wandb":
        from .wandb_logger import WandbLogger

        return WandbLogger(log_path, project)
    elif logger == "tensorboard":
        from .tensorboard_logger import TBLogger

        return TBLogger(log_path)
    else:
        return Logger(log_path)



import torch
import torch.nn as nn
import torch.nn.functional as F

from cogdl.utils.link_prediction_utils import GNNLinkPredict, sampling_edge_uniform
from cogdl.layers import RGCNLayer
from .. import BaseModel


class RGCN(nn.Module):
    def __init__(
        self,
        in_feats,
        out_feats,
        num_layers,
        num_rels,
        regularizer="basis",
        num_bases=None,
        self_loop=True,
        dropout=0.0,
        self_dropout=0.0,
    ):
        super(RGCN, self).__init__()
        shapes = [in_feats] + [out_feats] * num_layers
        self.num_layers = num_layers
        self.layers = nn.ModuleList(
            RGCNLayer(shapes[i], shapes[i + 1], num_rels, regularizer, num_bases, self_loop, dropout, self_dropout)
            for i in range(num_layers)
        )

    def forward(self, graph, x):
        h = x
        for i in range(len(self.layers)):
            h = self.layers[i](graph, h)
            if i < self.num_layers - 1:
                h = F.relu(h)
        return h


class LinkPredictRGCN(GNNLinkPredict, BaseModel):
    @staticmethod
    def add_args(parser):
        # fmt: off
        parser.add_argument("--hidden-size", type=int, default=200)
        parser.add_argument("--num-layers", type=int, default=2)
        parser.add_argument("--regularizer", type=str, default="basis")
        parser.add_argument("--self-loop", action="store_false")
        parser.add_argument("--penalty", type=float, default=0.001)
        parser.add_argument("--dropout", type=float, default=0.2)
        parser.add_argument("--self-dropout", type=float, default=0.4)
        parser.add_argument("--num-bases", type=int, default=5)
        parser.add_argument("--sampling-rate", type=float, default=0.01)
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            num_entities=args.num_entities,
            num_rels=args.num_rels,
            hidden_size=args.hidden_size,
            num_layers=args.num_layers,
            regularizer=args.regularizer,
            num_bases=args.num_bases,
            self_loop=args.self_loop,
            sampling_rate=args.sampling_rate,
            penalty=args.penalty,
            dropout=args.dropout,
            self_dropout=args.self_dropout,
        )

    def __init__(
        self,
        num_entities,
        num_rels,
        hidden_size,
        num_layers,
        regularizer="basis",
        num_bases=None,
        self_loop=True,
        sampling_rate=0.01,
        penalty=0,
        dropout=0.0,
        self_dropout=0.0,
    ):
        BaseModel.__init__(self)
        GNNLinkPredict.__init__(self)
        self.penalty = penalty
        self.num_nodes = num_entities
        self.num_rels = num_rels
        self.sampling_rate = sampling_rate
        self.edge_set = None

        self.model = RGCN(
            in_feats=hidden_size,
            out_feats=hidden_size,
            num_layers=num_layers,
            num_rels=num_rels,
            regularizer=regularizer,
            num_bases=num_bases,
            self_loop=self_loop,
            dropout=dropout,
            self_dropout=self_dropout,
        )
        # self.rel_weight = nn.Parameter(torch.Tensor(num_rels, hidden_size))
        # nn.init.xavier_normal_(self.rel_weight, gain=nn.init.calculate_gain("relu"))
        # self.emb = nn.Parameter(torch.Tensor(num_entities, hidden_size))
        # nn.init.xavier_normal_(self.emb, gain=nn.init.calculate_gain("relu"))

        self.rel_weight = nn.Embedding(num_rels, hidden_size)
        self.emb = nn.Embedding(num_entities, hidden_size)

    def forward(self, graph):
        reindexed_nodes, reindexed_edges = torch.unique(torch.stack(graph.edge_index), sorted=True, return_inverse=True)
        x = self.emb(reindexed_nodes)
        self.cahce_index = reindexed_nodes

        graph.edge_index = reindexed_edges
        # graph.num_nodes = reindexed_edges.max().item() + 1

        output = self.model(graph, x)
        # output = self.model(x, reindexed_indices, graph.edge_type)
        return output

    def loss(self, graph, scoring):
        edge_index = graph.edge_index
        edge_types = graph.edge_attr

        self.get_edge_set(edge_index, edge_types)
        batch_edges, batch_attr, samples, rels, labels = sampling_edge_uniform(
            edge_index, edge_types, self.edge_set, self.sampling_rate, self.num_rels
        )

        graph = graph.__class__(edge_index=batch_edges, edge_attr=batch_attr)
        # graph.edge_index = batch_edges
        # graph.edge_attr = batch_attr

        output = self.forward(graph)
        edge_weight = self.rel_weight(rels)
        sampled_nodes, reindexed_edges = torch.unique(samples, sorted=True, return_inverse=True)
        assert (sampled_nodes == self.cahce_index).any()
        sampled_types = torch.unique(rels)

        loss_n = self._loss(
            output[reindexed_edges[0]], output[reindexed_edges[1]], edge_weight, labels, scoring
        ) + self.penalty * self._regularization([self.emb(sampled_nodes), self.rel_weight(sampled_types)])
        return loss_n

    def predict(self, graph):
        device = next(self.parameters()).device
        indices = torch.arange(0, self.num_nodes).to(device)
        x = self.emb(indices)
        output = self.model(graph, x)

        return output, self.rel_weight.weight



from typing import List
import os
import random
import numpy as np
import scipy.sparse as sp
import torch
import torch.utils.data

from cogdl.utils import remove_self_loops, row_normalization
from cogdl.data import Graph, DataLoader


class NeighborSampler(DataLoader):
    def __init__(self, dataset, sizes: List[int], mask=None, **kwargs):
        if "batch_size" in kwargs:
            batch_size = kwargs["batch_size"]
        else:
            batch_size = 8

        if isinstance(dataset.data, Graph):
            self.dataset = NeighborSamplerDataset(dataset, sizes, batch_size, mask)
        else:
            self.dataset = dataset
        kwargs["batch_size"] = 1
        kwargs["shuffle"] = False
        kwargs["collate_fn"] = NeighborSampler.collate_fn
        super(NeighborSampler, self).__init__(dataset=self.dataset, **kwargs)

    @staticmethod
    def collate_fn(data):
        return data[0]

    def shuffle(self):
        self.dataset.shuffle()


class NeighborSamplerDataset(torch.utils.data.Dataset):
    def __init__(self, dataset, sizes: List[int], batch_size: int, mask=None):
        super(NeighborSamplerDataset, self).__init__()
        self.data = dataset.data
        self.x = self.data.x
        self.y = self.data.y
        self.sizes = sizes
        self.batch_size = batch_size
        self.node_idx = torch.arange(0, self.data.x.shape[0], dtype=torch.long)
        if mask is not None:
            self.node_idx = self.node_idx[mask]
        self.num_nodes = self.node_idx.shape[0]

    def shuffle(self):
        idx = torch.randperm(self.num_nodes)
        self.node_idx = self.node_idx[idx]

    def __len__(self):
        return (self.num_nodes - 1) // self.batch_size + 1

    def __getitem__(self, idx):
        """
            Sample a subgraph with neighborhood sampling
        Args:
            idx: torch.Tensor / np.array
                Target nodes
        Returns:
            if `size` is `[-1,]`,
                (
                    source_nodes_id: Tensor,
                    sampled_edges: Tensor,
                    (number_of_source_nodes, number_of_target_nodes): Tuple[int]
                )
            otherwise,
                (
                    target_nodes_id: Tensor
                    all_sampled_nodes_id: Tensor,
                    sampled_adjs: List[Tuple(Tensor, Tensor, Tuple[int]]
                )
        """
        batch = self.node_idx[idx * self.batch_size : (idx + 1) * self.batch_size]
        node_id = batch
        adj_list = []
        for size in self.sizes:
            src_id, graph = self.data.sample_adj(node_id, size, replace=False)
            size = (len(src_id), len(node_id))
            adj_list.append((src_id, graph, size))  # src_id, graph, (src_size, target_size)
            node_id = src_id

        if self.sizes == [-1]:
            src_id, graph, _ = adj_list[0]
            size = (len(src_id), len(batch))
            return src_id, graph, size
        else:
            return batch, node_id, adj_list[::-1]


class ClusteredDataset(torch.utils.data.Dataset):
    partition_tool = None

    def __init__(self, dataset, n_cluster: int, batch_size: int):
        super(ClusteredDataset).__init__()
        try:
            import metis

            ClusteredDataset.partition_tool = metis
        except Exception as e:
            print(e)
            exit(1)

        self.data = dataset.data
        self.dataset_name = dataset.__class__.__name__
        self.batch_size = batch_size
        self.n_cluster = n_cluster
        self.clusters = self.preprocess(n_cluster)
        self.batch_idx = np.array(range(n_cluster))

    def shuffle(self):
        random.shuffle(self.batch_idx)

    def __len__(self):
        return (self.n_cluster - 1) // self.batch_size + 1

    def __getitem__(self, idx):
        batch = self.batch_idx[idx * self.batch_size : (idx + 1) * self.batch_size]
        nodes = np.concatenate([self.clusters[i] for i in batch])
        subgraph = self.data.subgraph(nodes)
        subgraph.batch = torch.from_numpy(nodes)
        return subgraph

    def preprocess(self, n_cluster):
        save_name = f"{self.dataset_name}-{n_cluster}.cluster"
        if os.path.exists(save_name):
            return torch.load(save_name)
        print("Preprocessing...")
        row, col = self.data.edge_index
        (row, col), _ = remove_self_loops((row, col))
        if str(row.device) != "cpu":
            row = row.cpu().numpy()
            col = col.cpu().numpy()
        num_nodes = max(row.max(), col.max()) + 1
        adj = sp.csr_matrix((np.ones(row.shape[0]), (row, col)), shape=(num_nodes, num_nodes))
        indptr = adj.indptr
        indptr = np.split(adj.indices, indptr[1:])[:-1]
        _, parts = ClusteredDataset.partition_tool.part_graph(indptr, n_cluster, seed=1)
        division = [[] for _ in range(n_cluster)]
        for i, v in enumerate(parts):
            division[v].append(i)
        for k in range(len(division)):
            division[k] = np.array(division[k], dtype=np.int)
        torch.save(division, save_name)
        print("Graph clustering done")
        return division


class ClusteredLoader(DataLoader):
    def __init__(self, dataset, n_cluster: int, method="metis", **kwargs):
        if "batch_size" in kwargs:
            batch_size = kwargs["batch_size"]
        else:
            batch_size = 20

        if isinstance(dataset, ClusteredDataset) or isinstance(dataset, RandomPartitionDataset):
            self.dataset = dataset
        elif isinstance(dataset.data, Graph):
            if method == "metis":
                self.dataset = ClusteredDataset(dataset, n_cluster, batch_size)
            else:
                self.dataset = RandomPartitionDataset(dataset, n_cluster)
        kwargs["batch_size"] = 1
        kwargs["shuffle"] = False
        super(ClusteredLoader, self).__init__(dataset=self.dataset, collate_fn=ClusteredLoader.collate_fn, **kwargs)

    @staticmethod
    def collate_fn(item):
        return item[0]

    def shuffle(self):
        self.dataset.shuffle()


class RandomPartitionDataset(torch.utils.data.Dataset):
    """
    For ClusteredLoader
    """

    def __init__(self, dataset, n_cluster):
        self.data = dataset.data
        self.n_cluster = n_cluster
        self.num_nodes = dataset.data.num_nodes
        self.parts = torch.randint(0, self.n_cluster, size=(self.num_nodes,))

    def __getitem__(self, idx):
        node_cluster = torch.where(self.parts == idx)[0]
        subgraph = self.data.subgraph(node_cluster)
        subgraph.batch = node_cluster
        return subgraph

    def __len__(self):
        return self.n_cluster

    def shuffle(self):
        self.parts = torch.randint(0, self.n_cluster, size=(self.num_nodes,))



import torch
from cogdl.data import DataLoader

from cogdl.wrappers.data_wrapper.base_data_wrapper import OnLoadingWrapper


class DataController(object):
    def __init__(self, world_size: int = 1, distributed: bool = False):
        self.world_size = world_size
        self.distributed = distributed

    def distributed_dataloader(self, dataloader: DataLoader, dataset, rank):
        # TODO: just a toy implementation
        assert isinstance(dataloader, DataLoader)

        args, kwargs = dataloader.get_parameters()
        sampler = torch.utils.data.distributed.DistributedSampler(dataset, num_replicas=self.world_size, rank=rank)
        kwargs["sampler"] = sampler
        dataloader = dataloader.__class__(*args, **kwargs)
        return dataloader

    def prepare_data_wrapper(self, dataset_w, rank=0):
        if self.distributed:
            dataset_w.pre_transform()
            train_loader = dataset_w.train_wrapper()
            assert isinstance(train_loader, DataLoader)
            train_loader = self.distributed_dataloader(train_loader, dataset=dataset_w.get_train_dataset(), rank=rank)
            train_wrapper = OnLoadingWrapper(train_loader, dataset_w.train_transform)
            dataset_w.prepare_val_data()
            dataset_w.prepare_test_data()
            dataset_w.set_train_data(train_wrapper)
            return dataset_w
        else:
            dataset_w.pre_transform()
            dataset_w.prepare_training_data()
            dataset_w.prepare_val_data()
            dataset_w.prepare_test_data()
            return dataset_w

    def training_proc_per_stage(self, dataset_w, rank=0):
        if dataset_w.__refresh_per_epoch__():
            if self.distributed:
                train_loader = dataset_w.train_wrapper()
                assert isinstance(train_loader, DataLoader)
                train_loader = self.distributed_dataloader(train_loader, dataset=dataset_w.get_dataset(), rank=rank)
                train_wrapper = OnLoadingWrapper(train_loader, dataset_w.train_transform)
                dataset_w.__train_data = train_wrapper
            else:
                dataset_w.prepare_training_data()
        return dataset_w



import math

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.nn.modules.module import Module
from torch.nn.parameter import Parameter

from .. import BaseModel
from cogdl.utils import spmm


class GraphConvolutionBS(Module):
    """
    GCN Layer with BN, Self-loop and Res connection.
    """

    def __init__(
        self,
        in_features,
        out_features,
        activation=lambda x: x,
        withbn=True,
        withloop=True,
        bias=True,
        res=False,
    ):
        """
        Initial function.
        :param in_features: the input feature dimension.
        :param out_features: the output feature dimension.
        :param activation: the activation function.
        :param withbn: using batch normalization.
        :param withloop: using self feature modeling.
        :param bias: enable bias.
        :param res: enable res connections.
        """
        super(GraphConvolutionBS, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.sigma = activation
        self.res = res

        # Parameter setting.
        self.weight = Parameter(torch.FloatTensor(in_features, out_features))
        self.self_weight = Parameter(torch.FloatTensor(in_features, out_features)) if withloop else None

        self.bn = torch.nn.BatchNorm1d(out_features) if withbn else None
        self.bias = Parameter(torch.FloatTensor(out_features)) if bias else None

        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.self_weight is not None:
            stdv = 1.0 / math.sqrt(self.self_weight.size(1))
            self.self_weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, graph, x):
        support = torch.mm(x, self.weight)
        output = spmm(graph, support)

        # Self-loop
        output = output + torch.mm(x, self.self_weight) if self.self_weight is not None else output

        output = output + self.bias if self.bias is not None else output
        # BN
        output = self.bn(output) if self.bn is not None else output
        # Res
        return self.sigma(output) + input if self.res else self.sigma(output)

    def __repr__(self):
        return self.__class__.__name__ + " (" + str(self.in_features) + " -> " + str(self.out_features) + ")"


class GraphBaseBlock(Module):
    """
    The base block for Multi-layer GCN / ResGCN / Dense GCN
    """

    def __init__(
        self,
        in_features,
        out_features,
        nbaselayer,
        withbn=True,
        withloop=True,
        activation=F.relu,
        dropout=True,
        aggrmethod="concat",
        dense=False,
    ):
        """
        The base block for constructing DeepGCN model.
        :param in_features: the input feature dimension.
        :param out_features: the hidden feature dimension.
        :param nbaselayer: the number of layers in the base block.
        :param withbn: using batch normalization in graph convolution.
        :param withloop: using self feature modeling in graph convolution.
        :param activation: the activation function, default is ReLu.
        :param dropout: the dropout ratio.
        :param aggrmethod: the aggregation function for baseblock, can be "concat" and "add". For "resgcn", the default
                           is "add", for others the default is "concat".
        :param dense: enable dense connection
        """
        super(GraphBaseBlock, self).__init__()
        self.in_features = in_features
        self.hiddendim = out_features
        self.nhiddenlayer = nbaselayer
        self.activation = activation
        self.aggrmethod = aggrmethod
        self.dense = dense
        self.dropout = dropout
        self.withbn = withbn
        self.withloop = withloop
        self.hiddenlayers = nn.ModuleList()
        self.__makehidden()

        if self.aggrmethod == "concat" and dense is False:
            self.out_features = in_features + out_features
        elif self.aggrmethod == "concat" and dense is True:
            self.out_features = in_features + out_features * nbaselayer
        elif self.aggrmethod == "add":
            if in_features != self.hiddendim:
                raise RuntimeError("The dimension of in_features and hiddendim should be matched in add model.")
            self.out_features = out_features
        elif self.aggrmethod == "nores":
            self.out_features = out_features
        else:
            raise NotImplementedError("The aggregation method only support 'concat','add' and 'nores'.")

    def __makehidden(self):
        # for i in xrange(self.nhiddenlayer):
        for i in range(self.nhiddenlayer):
            if i == 0:
                layer = GraphConvolutionBS(
                    self.in_features,
                    self.hiddendim,
                    self.activation,
                    self.withbn,
                    self.withloop,
                )
            else:
                layer = GraphConvolutionBS(
                    self.hiddendim,
                    self.hiddendim,
                    self.activation,
                    self.withbn,
                    self.withloop,
                )
            self.hiddenlayers.append(layer)

    def _doconcat(self, x, subx):
        if x is None:
            return subx
        if self.aggrmethod == "concat":
            return torch.cat((x, subx), 1)
        elif self.aggrmethod == "add":
            return x + subx
        elif self.aggrmethod == "nores":
            return x

    def forward(self, graph, x):
        h = x
        denseout = None
        # Here out is the result in all levels.
        for gc in self.hiddenlayers:
            denseout = self._doconcat(denseout, h)
            h = gc(graph, h)
            h = F.dropout(h, self.dropout, training=self.training)

        if not self.dense:
            return self._doconcat(h, x)

        return self._doconcat(h, denseout) if denseout is not None else h

    def get_outdim(self):
        return self.out_features

    def __repr__(self):
        return "%s %s (%d - [%d:%d] > %d)" % (
            self.__class__.__name__,
            self.aggrmethod,
            self.in_features,
            self.hiddendim,
            self.nhiddenlayer,
            self.out_features,
        )


class MultiLayerGCNBlock(Module):
    """
    Muti-Layer GCN with same hidden dimension.
    """

    def __init__(
        self,
        in_features,
        out_features,
        nbaselayer,
        withbn=True,
        withloop=True,
        activation=F.relu,
        dropout=True,
        aggrmethod=None,
        dense=None,
    ):
        """
        The multiple layer GCN block.
        :param in_features: the input feature dimension.
        :param out_features: the hidden feature dimension.
        :param nbaselayer: the number of layers in the base block.
        :param withbn: using batch normalization in graph convolution.
        :param withloop: using self feature modeling in graph convolution.
        :param activation: the activation function, default is ReLu.
        :param dropout: the dropout ratio.
        :param aggrmethod: not applied.
        :param dense: not applied.
        """
        super(MultiLayerGCNBlock, self).__init__()
        self.model = GraphBaseBlock(
            in_features=in_features,
            out_features=out_features,
            nbaselayer=nbaselayer,
            withbn=withbn,
            withloop=withloop,
            activation=activation,
            dropout=dropout,
            dense=False,
            aggrmethod="nores",
        )

    def forward(self, graph, x):
        return self.model.forward(graph, x)

    def get_outdim(self):
        return self.model.get_outdim()

    def __repr__(self):
        return "%s %s (%d - [%d:%d] > %d)" % (
            self.__class__.__name__,
            self.aggrmethod,
            self.model.in_features,
            self.model.hiddendim,
            self.model.nhiddenlayer,
            self.model.out_features,
        )


class ResGCNBlock(Module):
    """
    The multiple layer GCN with residual connection block.
    """

    def __init__(
        self,
        in_features,
        out_features,
        nbaselayer,
        withbn=True,
        withloop=True,
        activation=F.relu,
        dropout=True,
        aggrmethod=None,
        dense=None,
    ):
        """
        The multiple layer GCN with residual connection block.
        :param in_features: the input feature dimension.
        :param out_features: the hidden feature dimension.
        :param nbaselayer: the number of layers in the base block.
        :param withbn: using batch normalization in graph convolution.
        :param withloop: using self feature modeling in graph convolution.
        :param activation: the activation function, default is ReLu.
        :param dropout: the dropout ratio.
        :param aggrmethod: not applied.
        :param dense: not applied.
        """
        super(ResGCNBlock, self).__init__()
        self.model = GraphBaseBlock(
            in_features=in_features,
            out_features=out_features,
            nbaselayer=nbaselayer,
            withbn=withbn,
            withloop=withloop,
            activation=activation,
            dropout=dropout,
            dense=False,
            aggrmethod="add",
        )

    def forward(self, graph, x):
        return self.model.forward(graph, x)

    def get_outdim(self):
        return self.model.get_outdim()

    def __repr__(self):
        return "%s %s (%d - [%d:%d] > %d)" % (
            self.__class__.__name__,
            self.aggrmethod,
            self.model.in_features,
            self.model.hiddendim,
            self.model.nhiddenlayer,
            self.model.out_features,
        )


class DenseGCNBlock(Module):
    """
    The multiple layer GCN with dense connection block.
    """

    def __init__(
        self,
        in_features,
        out_features,
        nbaselayer,
        withbn=True,
        withloop=True,
        activation=F.relu,
        dropout=True,
        aggrmethod="concat",
        dense=True,
    ):
        """
        The multiple layer GCN with dense connection block.
        :param in_features: the input feature dimension.
        :param out_features: the hidden feature dimension.
        :param nbaselayer: the number of layers in the base block.
        :param withbn: using batch normalization in graph convolution.
        :param withloop: using self feature modeling in graph convolution.
        :param activation: the activation function, default is ReLu.
        :param dropout: the dropout ratio.
        :param aggrmethod: the aggregation function for the output. For denseblock, default is "concat".
        :param dense: default is True, cannot be changed.
        """
        super(DenseGCNBlock, self).__init__()
        self.model = GraphBaseBlock(
            in_features=in_features,
            out_features=out_features,
            nbaselayer=nbaselayer,
            withbn=withbn,
            withloop=withloop,
            activation=activation,
            dropout=dropout,
            dense=True,
            aggrmethod=aggrmethod,
        )

    def forward(self, graph, x):
        return self.model.forward(graph, x)

    def get_outdim(self):
        return self.model.get_outdim()

    def __repr__(self):
        return "%s %s (%d - [%d:%d] > %d)" % (
            self.__class__.__name__,
            self.aggrmethod,
            self.model.in_features,
            self.model.hiddendim,
            self.model.nhiddenlayer,
            self.model.out_features,
        )


class InceptionGCNBlock(Module):
    """
    The multiple layer GCN with inception connection block.
    """

    def __init__(
        self,
        in_features,
        out_features,
        nbaselayer,
        withbn=True,
        withloop=True,
        activation=F.relu,
        dropout=True,
        aggrmethod="concat",
        dense=False,
    ):
        """
        The multiple layer GCN with inception connection block.
        :param in_features: the input feature dimension.
        :param out_features: the hidden feature dimension.
        :param nbaselayer: the number of layers in the base block.
        :param withbn: using batch normalization in graph convolution.
        :param withloop: using self feature modeling in graph convolution.
        :param activation: the activation function, default is ReLu.
        :param dropout: the dropout ratio.
        :param aggrmethod: the aggregation function for baseblock, can be "concat" and "add". For "resgcn", the default
                           is "add", for others the default is "concat".
        :param dense: not applied. The default is False, cannot be changed.
        """
        super(InceptionGCNBlock, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.hiddendim = out_features
        self.nbaselayer = nbaselayer
        self.activation = activation
        self.aggrmethod = aggrmethod
        self.dropout = dropout
        self.withbn = withbn
        self.withloop = withloop
        self.midlayers = nn.ModuleList()
        self.__makehidden()

        if self.aggrmethod == "concat":
            self.out_features = in_features + out_features * nbaselayer
        elif self.aggrmethod == "add":
            if in_features != self.hiddendim:
                raise RuntimeError("The dimension of in_features and hiddendim should be matched in 'add' model.")
            self.out_features = out_features
        else:
            raise NotImplementedError("The aggregation method only support 'concat', 'add'.")

    def __makehidden(self):
        # for j in xrange(self.nhiddenlayer):
        for j in range(self.nbaselayer):
            reslayer = nn.ModuleList()
            # for i in xrange(j + 1):
            for i in range(j + 1):
                if i == 0:
                    layer = GraphConvolutionBS(
                        self.in_features,
                        self.hiddendim,
                        self.activation,
                        self.withbn,
                        self.withloop,
                    )
                else:
                    layer = GraphConvolutionBS(
                        self.hiddendim,
                        self.hiddendim,
                        self.activation,
                        self.withbn,
                        self.withloop,
                    )
                reslayer.append(layer)
            self.midlayers.append(reslayer)

    def forward(self, graph, x):
        for reslayer in self.midlayers:
            subx = x
            for gc in reslayer:
                subx = gc(graph, x)
                subx = F.dropout(subx, self.dropout, training=self.training)
            x = self._doconcat(x, subx)
        return x

    def get_outdim(self):
        return self.out_features

    def _doconcat(self, x, subx):
        if self.aggrmethod == "concat":
            return torch.cat((x, subx), 1)
        elif self.aggrmethod == "add":
            return x + subx

    def __repr__(self):
        return "%s %s (%d - [%d:%d] > %d)" % (
            self.__class__.__name__,
            self.aggrmethod,
            self.in_features,
            self.hiddendim,
            self.nbaselayer,
            self.out_features,
        )


class Dense(Module):
    """
    Simple Dense layer, Do not consider adj.
    """

    def __init__(self, in_features, out_features, activation=lambda x: x, bias=True, res=False):
        super(Dense, self).__init__()
        self.in_features = in_features
        self.out_features = out_features
        self.sigma = activation
        self.weight = Parameter(torch.FloatTensor(in_features, out_features))
        self.res = res
        self.bn = nn.BatchNorm1d(out_features)
        self.bias = Parameter(torch.FloatTensor(out_features)) if bias else None
        self.reset_parameters()

    def reset_parameters(self):
        stdv = 1.0 / math.sqrt(self.weight.size(1))
        self.weight.data.uniform_(-stdv, stdv)
        if self.bias is not None:
            self.bias.data.uniform_(-stdv, stdv)

    def forward(self, graph, x):
        output = torch.mm(x, self.weight)
        output = output + self.bias if self.bias is not None else output
        output = self.bn(output)
        return self.sigma(output)

    def __repr__(self):
        return self.__class__.__name__ + " (" + str(self.in_features) + " -> " + str(self.out_features) + ")"


class DropEdge_GCN(BaseModel):
    """
     DropEdge: Towards Deep Graph Convolutional Networks on Node Classification
     Applying DropEdge to GCN @ https://arxiv.org/pdf/1907.10903.pdf

    The model for the single kind of deepgcn blocks.
    The model architecture likes:
    inputlayer(nfeat)--block(nbaselayer, nhid)--...--outputlayer(nclass)--softmax(nclass)
                        |------  nhidlayer  ----|
    The total layer is nhidlayer*nbaselayer + 2.
    All options are configurable.

     Args:
         Initial function.
         :param nfeat: the input feature dimension.
         :param nhid:  the hidden feature dimension.
         :param nclass: the output feature dimension.
         :param nhidlayer: the number of hidden blocks.
         :param dropout:  the dropout ratio.
         :param baseblock: the baseblock type, can be "mutigcn", "resgcn", "densegcn" and "inceptiongcn".
         :param inputlayer: the input layer type, can be "gcn", "dense", "none".
         :param outputlayer: the input layer type, can be "gcn", "dense".
         :param nbaselayer: the number of layers in one hidden block.
         :param activation: the activation function, default is ReLu.
         :param withbn: using batch normalization in graph convolution.
         :param withloop: using self feature modeling in graph convolution.
         :param aggrmethod: the aggregation function for baseblock, can be "concat" and "add". For "resgcn", the default
                            is "add", for others the default is "concat".
    """

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--num_features", type=int)

        parser.add_argument("--num_classes", type=int)

        parser.add_argument('--baseblock', default='mutigcn',
                            help="Choose the model to be trained.( mutigcn, resgcn, densegcn, inceptiongcn)")
        parser.add_argument('--inputlayer', default='gcn',
                            help="The input layer of the model.")
        parser.add_argument('--outputlayer', default='gcn',
                            help="The output layer of the model.")
        parser.add_argument('--hidden-size', type=int, default=64,
                            help='Number of hidden units.')
        parser.add_argument('--dropout', type=float, default=0.5,
                            help='Dropout rate (1 - keep probability).')
        parser.add_argument('--withbn', action='store_true', default=False,
                            help='Enable Bath Norm GCN')
        parser.add_argument('--withloop', action="store_true", default=False,
                            help="Enable loop layer GCN")
        parser.add_argument('--nhiddenlayer', type=int, default=1,
                            help='The number of hidden layers.')

        parser.add_argument("--nbaseblocklayer", type=int, default=0,
                            help="The number of layers in each baseblock")
        parser.add_argument("--aggrmethod", default="default",
                            help="The aggrmethod for the layer aggreation. The options includes add and concat. Only valid in resgcn, densegcn and inceptiongcn")
        parser.add_argument("--task_type", default="full", help="The node classification task type (full and semi). Only valid for cora, citeseer and pubmed dataset.")

        parser.add_argument("--activation", default=F.relu, help="activiation function")

        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.hidden_size,
            args.num_classes,
            args.nhiddenlayer,
            args.dropout,
            args.baseblock,
            args.inputlayer,
            args.outputlayer,
            args.nbaseblocklayer,
            args.activation,
            args.withbn,
            args.withloop,
            args.aggrmethod,
        )

    def __init__(
        self,
        nfeat,
        nhid,
        nclass,
        nhidlayer,
        dropout,
        baseblock,
        inputlayer,
        outputlayer,
        nbaselayer,
        activation,
        withbn,
        withloop,
        aggrmethod,
    ):
        super(DropEdge_GCN, self).__init__()

        self.dropout = dropout

        if baseblock == "resgcn":
            self.BASEBLOCK = ResGCNBlock
        elif baseblock == "densegcn":
            self.BASEBLOCK = DenseGCNBlock
        elif baseblock == "mutigcn":
            self.BASEBLOCK = MultiLayerGCNBlock
        elif baseblock == "inceptiongcn":
            self.BASEBLOCK = InceptionGCNBlock
        else:
            raise NotImplementedError("Current baseblock %s is not supported." % (baseblock))
        if inputlayer == "gcn":
            # input gc
            self.ingc = GraphConvolutionBS(nfeat, nhid, activation, withbn, withloop)
            baseblockinput = nhid
        elif inputlayer == "none":
            self.ingc = lambda x: x
            baseblockinput = nfeat
        else:
            self.ingc = Dense(nfeat, nhid, activation)
            baseblockinput = nhid

        outactivation = lambda x: x  # noqa E731
        if outputlayer == "gcn":
            self.outgc = GraphConvolutionBS(baseblockinput, nclass, outactivation, withbn, withloop)
        # elif outputlayer ==  "none": #here can not be none
        #    self.outgc = lambda x: x
        else:
            self.outgc = Dense(nhid, nclass, activation)

        # hidden layer
        self.midlayer = nn.ModuleList()
        # Dense is not supported now.
        # for i in xrange(nhidlayer):
        for i in range(nhidlayer):
            gcb = self.BASEBLOCK(
                in_features=baseblockinput,
                out_features=nhid,
                nbaselayer=nbaselayer,
                withbn=withbn,
                withloop=withloop,
                activation=activation,
                dropout=dropout,
                dense=False,
                aggrmethod=aggrmethod,
            )
            self.midlayer.append(gcb)
            baseblockinput = gcb.get_outdim()
        # output gc
        outactivation = lambda x: x  # noqa E731 we donot need nonlinear activation here.
        self.outgc = GraphConvolutionBS(baseblockinput, nclass, outactivation, withbn, withloop)

        self.reset_parameters()

    def reset_parameters(self):
        pass

    def forward(self, graph):
        x = graph.x

        x = self.ingc(graph, x)
        x = F.dropout(x, self.dropout, training=self.training)

        # mid block connections
        # for i in xrange(len(self.midlayer)):
        for i in range(len(self.midlayer)):
            midgc = self.midlayer[i]
            x = midgc(graph, x)
        # output, no relu and dropput here.
        x = self.outgc(graph, x)
        x = F.log_softmax(x, dim=1)
        return x

    def predict(self, data):
        return self.forward(data)



from typing import Optional

import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.utils.checkpoint import checkpoint

from .mlp_layer import MLP
from cogdl.utils import get_activation, edge_softmax, get_norm_layer, batch_max_pooling


class GENConv(nn.Module):
    def __init__(
        self,
        in_feats: int,
        out_feats: int,
        aggr: str = "softmax_sg",
        beta: float = 1.0,
        p: float = 1.0,
        learn_beta: bool = False,
        learn_p: bool = False,
        use_msg_norm: bool = False,
        learn_msg_scale: bool = True,
        norm: Optional[str] = None,
        residual: bool = False,
        activation: Optional[str] = None,
        num_mlp_layers: int = 2,
        edge_attr_size: Optional[list] = None,
    ):
        super(GENConv, self).__init__()
        self.use_msg_norm = use_msg_norm
        self.mlp = MLP(in_feats, out_feats, in_feats * 2, num_layers=num_mlp_layers, activation=activation, norm=norm)

        self.message_encoder = torch.nn.ReLU()

        self.aggr = aggr
        if aggr == "softmax_sg":
            self.beta = torch.nn.Parameter(
                torch.Tensor(
                    [
                        beta,
                    ]
                ),
                requires_grad=learn_beta,
            )
        else:
            self.register_buffer("beta", None)
        if aggr == "powermean":
            self.p = torch.nn.Parameter(
                torch.Tensor(
                    [
                        p,
                    ]
                ),
                requires_grad=learn_p,
            )
        else:
            self.register_buffer("p", None)
        self.eps = 1e-7

        self.s = torch.nn.Parameter(torch.Tensor([1.0]), requires_grad=learn_msg_scale and use_msg_norm)
        self.residual = residual

        if edge_attr_size is not None and edge_attr_size[0] > 0:
            if len(edge_attr_size) > 1:
                self.edge_encoder = BondEncoder(edge_attr_size, in_feats)
            else:
                self.edge_encoder = EdgeEncoder(edge_attr_size[0], in_feats)
        else:
            self.edge_encoder = None

    def message_norm(self, x, msg):
        x_norm = torch.norm(x, dim=1, p=2)
        msg_norm = F.normalize(msg, p=2, dim=1)
        msg_norm = msg_norm * x_norm.unsqueeze(-1)
        return x + self.s * msg_norm

    def forward(self, graph, x):
        edge_index = graph.edge_index
        dim = x.shape[1]
        edge_msg = x[edge_index[1]]
        if self.edge_encoder is not None and graph.edge_attr is not None:
            edge_msg += self.edge_encoder(graph.edge_attr)
        edge_msg = self.message_encoder(edge_msg) + self.eps

        if self.aggr == "softmax_sg":
            h = edge_softmax(graph, self.beta * edge_msg.contiguous())
            h = edge_msg * h
        elif self.aggr == "softmax":
            h = edge_softmax(graph, edge_msg)
            h = edge_msg * h
        elif self.aggr == "powermean":
            deg = graph.degrees()
            torch.clamp_(edge_msg, 1e-7, 1.0)
            h = edge_msg.pow(self.p) / deg[edge_index[0]].unsqueeze(-1)
        elif self.aggr == "mean":
            deg = graph.degrees()
            deg_rev = deg.pow(-1)
            deg_rev[torch.isinf(deg_rev)] = 0
            h = edge_msg * deg_rev[edge_index[0]].unsqueeze(-1)
        else:
            h = edge_msg

        if self.aggr == "max":
            h = batch_max_pooling(h, edge_index[0])
        else:
            h = torch.zeros_like(x).scatter_add_(dim=0, index=edge_index[0].unsqueeze(-1).repeat(1, dim), src=h)
        if self.aggr == "powermean":
            h = h.pow(1.0 / self.p)
        if self.use_msg_norm:
            h = self.message_norm(x, h)

        if self.residual:
            h = h + x
        h = self.mlp(h)
        return h


class ResGNNLayer(nn.Module):
    """
    Implementation of DeeperGCN in paper `"DeeperGCN: All You Need to Train Deeper GCNs" <https://arxiv.org/abs/2006.07739>`_

    Parameters
    -----------
    conv : nn.Module
        An instance of GNN Layer, recieving (graph, x) as inputs
    n_channels : int
        size of input features
    activation : str
    norm: str
        type of normalization, ``batchnorm`` as default
    dropout : float
    checkpoint_grad : bool
    """

    def __init__(
        self,
        conv,
        in_channels,
        activation="relu",
        norm="batchnorm",
        dropout=0.0,
        out_norm=None,
        out_channels=-1,
        residual=True,
        checkpoint_grad=False,
    ):
        super(ResGNNLayer, self).__init__()
        self.conv = conv
        self.activation = get_activation(activation)
        self.dropout = dropout
        self.norm = get_norm_layer(norm, in_channels)
        self.residual = residual
        if out_norm:
            self.out_norm = get_norm_layer(norm, out_channels)
        else:
            self.out_norm = None
        self.checkpoint_grad = checkpoint_grad

    def forward(self, graph, x, dropout=None, *args, **kwargs):
        h = self.norm(x)
        h = self.activation(h)
        if isinstance(dropout, float) or dropout is None:
            h = F.dropout(h, p=self.dropout, training=self.training)
        else:
            if self.training:
                h = h * dropout

        if self.checkpoint_grad:
            h = checkpoint(self.conv, graph, h, *args, **kwargs)
        else:
            h = self.conv(graph, h, *args, **kwargs)
        if self.residual:
            h = h + x

        if self.out_norm:
            return self.out_norm(h)
        else:
            return h


class EdgeEncoder(nn.Module):
    def __init__(self, in_feats, out_feats, bias=False):
        super(EdgeEncoder, self).__init__()
        self.nn = nn.Linear(in_feats, out_feats, bias=bias)

    def forward(self, edge_attr):
        return self.nn(edge_attr)


class BondEncoder(nn.Module):
    def __init__(self, bond_dim_list, emb_size):
        super(BondEncoder, self).__init__()
        self.bond_emb_list = nn.ModuleList()
        for i, size in enumerate(bond_dim_list):
            x = nn.Embedding(size, emb_size)
            self.bond_emb_list.append(x)

    def forward(self, edge_attr):
        out = 0
        for i in range(edge_attr.shape[1]):
            out += self.bond_emb_list[i](edge_attr[:, i])
        return out



import warnings
from . import Logger

try:
    import wandb
except Exception:
    warnings.warn("Please install wandb first")


class WandbLogger(Logger):
    def __init__(self, log_path, project=None):
        super(WandbLogger, self).__init__(log_path)
        self.last_step = 0
        self.project = project

    def start(self):
        self.run = wandb.init(reinit=True, dir=self.log_path, project=self.project)

    def note(self, metrics, step=None):
        if not hasattr(self, "run"):
            self.start()
        if step is None:
            step = self.last_step
        self.run.log(metrics, step=step)
        self.last_step = step

    def finish(self):
        self.run.finish()



from .heterogeneous_embedding_mw import HeterogeneousEmbeddingModelWrapper
from .heterogeneous_gnn_mw import HeterogeneousGNNModelWrapper
from .multiplex_embedding_mw import MultiplexEmbeddingModelWrapper



import torch
from torch import nn
import torch.nn.functional as F
import numpy as np

from .. import BaseModel
from cogdl.layers import SAINTLayer


def parse_arch(architecture, aggr, act, bias, hidden_size, num_features):
    num_layers = len(architecture.split("-"))
    # set default values, then update by arch_gcn
    bias_layer = [bias] * num_layers
    act_layer = [act] * num_layers
    aggr_layer = [aggr] * num_layers
    dims_layer = [hidden_size] * num_layers
    order_layer = [int(order) for order in architecture.split("-")]
    return [num_features] + dims_layer, order_layer, act_layer, bias_layer, aggr_layer


class GraphSAINT(BaseModel):
    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--num-features", type=int)
        parser.add_argument("--hidden-size", type=int, default=128)
        parser.add_argument("--architecture", type=str, default="1-1-0")
        parser.add_argument("--aggr", type=str, default="concat")
        parser.add_argument("--act", type=str, default="relu")
        parser.add_argument("--bias", type=str, default="norm")
        parser.add_argument("--dropout", type=float, default=0.1)
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.num_features,
            args.num_classes,
            args.architecture,
            args.aggr,
            args.act,
            args.bias,
            args.dropout,
            args.hidden_size,
        )

    def __init__(self, num_features, num_classes, architecture, aggr, act, bias, dropout, hidden_size):
        """
        Build the multi-layer GNN architecture.

        Inputs:
            num_classes         int, number of classes a node can belong to
            arch_gcn            dict, config for each GNN layer
            train_params        dict, training hyperparameters (e.g., learning rate)
            feat_full           np array of shape N x f, where N is the total num of
                                nodes and f is the dimension for input node feature
            label_full          np array, for single-class classification, the shape
                                is N x 1 and for multi-class classification, the
                                shape is N x c (where c = num_classes)
            cpu_eval            bool, if True, will put the model on CPU.

        Outputs:
            None
        """
        super(GraphSAINT, self).__init__()
        self.aggregator_cls = SAINTLayer
        self.mulhead = 1
        self.dropout = dropout
        self.sigmoid_loss = True
        self.num_classes = num_classes
        self.num_layers = len(architecture.split("-"))
        _dims, self.order_layer, self.act_layer, self.bias_layer, self.aggr_layer = parse_arch(
            architecture, aggr, act, bias, hidden_size, num_features
        )
        # get layer index for each conv layer, useful for jk net last layer aggregation
        self.set_idx_conv()
        self.set_dims(_dims)

        self.loss = 0
        self.opt_op = None

        # build the model below
        self.num_params = 0
        self.aggregators, num_param = self.get_aggregators()
        self.num_params += num_param
        self.conv_layers = nn.ModuleList(self.aggregators)
        self.classifier = SAINTLayer(
            self.dims_feat[-1], self.num_classes, act="I", order=0, dropout=self.dropout, bias="bias"
        )
        self.num_params += self.classifier.num_param

    def set_dims(self, dims):
        """
        Set the feature dimension / weight dimension for each GNN or MLP layer.
        We will use the dimensions set here to initialize PyTorch layers.

        Inputs:
            dims        list, length of node feature for each hidden layer

        Outputs:
            None
        """
        self.dims_feat = [dims[0]] + [
            ((self.aggr_layer[layer] == "concat") * self.order_layer[layer] + 1) * dims[layer + 1]
            for layer in range(len(dims) - 1)
        ]
        self.dims_weight = [(self.dims_feat[layer], dims[layer + 1]) for layer in range(len(dims) - 1)]

    def set_idx_conv(self):
        """
        Set the index of GNN layers for the full neural net. For example, if
        the full NN is having 1-0-1-0 arch (1-hop graph conv, followed by 0-hop
        MLP, ...). Then the layer indices will be 0, 2.
        """
        idx_conv = np.where(np.array(self.order_layer) >= 1)[0]
        idx_conv = list(idx_conv[1:] - 1)
        idx_conv.append(len(self.order_layer) - 1)
        _o_arr = np.array(self.order_layer)[idx_conv]
        if np.prod(np.ediff1d(_o_arr)) == 0:
            self.idx_conv = idx_conv
        else:
            self.idx_conv = list(np.where(np.array(self.order_layer) == 1)[0])

    def forward(self, graph):
        x = graph.x
        for layer in self.conv_layers:
            x = layer(graph, x)
        emb_subg_norm = F.normalize(x, p=2, dim=1)
        pred_subg = self.classifier(None, emb_subg_norm)
        return pred_subg

    def _loss(self, preds, labels, norm_loss):
        """
        The predictor performs sigmoid (for multi-class) or softmax (for single-class)
        """
        if self.sigmoid_loss:
            norm_loss = norm_loss.unsqueeze(1)
            return torch.nn.BCEWithLogitsLoss(weight=norm_loss, reduction="sum")(preds, labels)
        else:
            _ls = torch.nn.CrossEntropyLoss(reduction="none")(preds, labels)
            return (norm_loss * _ls).sum()

    def get_aggregators(self):
        """
        Return a list of aggregator instances. to be used in self.build()
        """
        num_param = 0
        aggregators = []
        for layer in range(self.num_layers):
            aggr = self.aggregator_cls(
                *self.dims_weight[layer],
                dropout=self.dropout,
                act=self.act_layer[layer],
                order=self.order_layer[layer],
                aggr=self.aggr_layer[layer],
                bias=self.bias_layer[layer],
                mulhead=self.mulhead,
            )
            num_param += aggr.num_param
            aggregators.append(aggr)
        return aggregators, num_param

    def predict(self, data):
        return self.forward(data)



import time

import networkx as nx
import numpy as np
import scipy.sparse as sp
import torch
import torch.nn as nn
import torch.nn.functional as F
from tqdm import tqdm

from .. import BaseModel


class SDNE_layer(nn.Module):
    def __init__(self, num_node, hidden_size1, hidden_size2, droput, alpha, beta, nu1, nu2):
        super(SDNE_layer, self).__init__()
        self.num_node = num_node
        self.hidden_size1 = hidden_size1
        self.hidden_size2 = hidden_size2
        self.droput = droput
        self.alpha = alpha
        self.beta = beta
        self.nu1 = nu1
        self.nu2 = nu2

        self.encode0 = nn.Linear(self.num_node, self.hidden_size1)
        self.encode1 = nn.Linear(self.hidden_size1, self.hidden_size2)
        self.decode0 = nn.Linear(self.hidden_size2, self.hidden_size1)
        self.decode1 = nn.Linear(self.hidden_size1, self.num_node)

    def forward(self, adj_mat, l_mat):
        t0 = F.leaky_relu(self.encode0(adj_mat))
        t0 = F.leaky_relu(self.encode1(t0))
        self.embedding = t0
        t0 = F.leaky_relu(self.decode0(t0))
        t0 = F.leaky_relu(self.decode1(t0))

        L_1st = 2 * torch.trace(torch.mm(torch.mm(torch.t(self.embedding), l_mat), self.embedding))
        L_2nd = torch.sum(((adj_mat - t0) * adj_mat * self.beta) * ((adj_mat - t0) * adj_mat * self.beta))

        L_reg = 0
        for param in self.parameters():
            L_reg += self.nu1 * torch.sum(torch.abs(param)) + self.nu2 * torch.sum(param * param)
        return self.alpha * L_1st, L_2nd, self.alpha * L_1st + L_2nd, L_reg

    def get_emb(self, adj):
        t0 = self.encode0(adj)
        t0 = self.encode1(t0)
        return t0


class SDNE(BaseModel):
    r"""The SDNE model from the `"Structural Deep Network Embedding"
    <https://www.kdd.org/kdd2016/papers/files/rfp0191-wangAemb.pdf>`_ paper

    Args:
        hidden_size1 (int) : The size of the first hidden layer.
        hidden_size2 (int) : The size of the second hidden layer.
        droput (float) : Droput rate.
        alpha (float) : Trade-off parameter between 1-st and 2-nd order objective function in SDNE.
        beta (float) : Parameter of 2-nd order objective function in SDNE.
        nu1 (float) : Parameter of l1 normlization in SDNE.
        nu2 (float) : Parameter of l2 normlization in SDNE.
        epochs (int) : The max epoches in training step.
        lr (float) : Learning rate in SDNE.
        cpu (bool) : Use CPU or GPU to train hin2vec.
    """

    @staticmethod
    def add_args(parser):
        """Add model-specific arguments to the parser."""
        # fmt: off
        parser.add_argument("--hidden-size1", type=int, default=1000, help="Hidden size in first layer of Auto-Encoder")
        parser.add_argument("--hidden-size2", type=int, default=128, help="Hidden size in second layer of Auto-Encoder")
        parser.add_argument("--droput", type=float, default=0.5, help="Dropout rate")
        parser.add_argument("--alpha", type=float, default=1e-1, help="alhpa is a hyperparameter in SDNE")
        parser.add_argument("--beta", type=float, default=5, help="beta is a hyperparameter in SDNE")
        parser.add_argument("--nu1", type=float, default=1e-4, help="nu1 is a hyperparameter in SDNE")
        parser.add_argument("--nu2", type=float, default=1e-3, help="nu2 is a hyperparameter in SDNE")
        # fmt: on

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            args.hidden_size1,
            args.hidden_size2,
            args.droput,
            args.alpha,
            args.beta,
            args.nu1,
            args.nu2,
            args.epochs,
            args.lr,
            args.cpu,
        )

    def __init__(self, hidden_size1, hidden_size2, droput, alpha, beta, nu1, nu2, epochs, lr, cpu):
        super(SDNE, self).__init__()
        self.hidden_size1 = hidden_size1
        self.hidden_size2 = hidden_size2
        self.droput = droput
        self.alpha = alpha
        self.beta = beta
        self.nu1 = nu1
        self.nu2 = nu2
        self.epochs = epochs
        self.lr = lr
        self.cpu = cpu

    def forward(self, graph, return_dict=False):
        device = "cuda" if torch.cuda.is_available() and not self.cpu else "cpu"
        G = graph.to_networkx()
        num_node = G.number_of_nodes()
        model = SDNE_layer(
            num_node, self.hidden_size1, self.hidden_size2, self.droput, self.alpha, self.beta, self.nu1, self.nu2
        )

        A = torch.from_numpy(nx.adjacency_matrix(G).todense().astype(np.float32))
        L = torch.from_numpy(nx.laplacian_matrix(G).todense().astype(np.float32))

        A, L = A.to(device), L.to(device)
        model = model.to(device)

        opt = torch.optim.Adam(model.parameters(), lr=self.lr)
        epoch_iter = tqdm(range(self.epochs))
        for epoch in epoch_iter:
            opt.zero_grad()
            L_1st, L_2nd, L_all, L_reg = model.forward(A, L)
            Loss = L_all + L_reg
            Loss.backward()
            epoch_iter.set_description(
                f"Epoch: {epoch:03d}, L_1st: {L_1st:.4f}, L_2nd: {L_2nd:.4f}, L_reg: {L_reg:.4f}"
            )
            opt.step()
        embeddings = model.get_emb(A)
        embeddings = embeddings.detach().cpu().numpy()

        if return_dict:
            features_matrix = dict()
            for vid, node in enumerate(G.nodes()):
                features_matrix[node] = embeddings[vid]
        else:
            features_matrix = np.zeros((graph.num_nodes, embeddings.shape[1]))
            nx_nodes = G.nodes()
            features_matrix[nx_nodes] = embeddings[np.arange(graph.num_nodes)]
        return features_matrix



import numba
import numpy as np
import scipy.sparse as sp
import torch


@numba.njit(cache=True, locals={"_val": numba.float32, "res": numba.float32, "res_vnode": numba.float32})
def _calc_ppr_node(inode, indptr, indices, deg, alpha, epsilon):
    alpha_eps = alpha * epsilon
    f32_0 = numba.float32(0)
    p = {inode: f32_0}
    r = {inode: alpha}
    q = [inode]
    while len(q) > 0:
        unode = q.pop()

        res = r[unode] if unode in r else f32_0
        if unode in p:
            p[unode] += res
        else:
            p[unode] = res
        r[unode] = f32_0
        for vnode in indices[indptr[unode] : indptr[unode + 1]]:
            _val = (1 - alpha) * res / deg[unode]
            if vnode in r:
                r[vnode] += _val
            else:
                r[vnode] = _val

            res_vnode = r[vnode] if vnode in r else f32_0
            if res_vnode >= alpha_eps * deg[vnode]:
                if vnode not in q:
                    q.append(vnode)

    return list(p.keys()), list(p.values())


@numba.njit(cache=True, parallel=True)
def calc_ppr_topk_parallel(indptr, indices, deg, alpha, epsilon, nodes, topk):
    js = [np.zeros(0, dtype=np.int64)] * len(nodes)
    vals = [np.zeros(0, dtype=np.float32)] * len(nodes)
    for i in numba.prange(len(nodes)):
        j, val = _calc_ppr_node(nodes[i], indptr, indices, deg, alpha, epsilon)
        j_np, val_np = np.array(j), np.array(val)
        idx_topk = np.argsort(val_np)[-topk:]
        js[i] = j_np[idx_topk]
        vals[i] = val_np[idx_topk]
    return js, vals


def ppr_topk(adj_matrix, alpha, epsilon, nodes, topk):
    """Calculate the PPR matrix approximately using Anderson."""

    out_degree = np.sum(adj_matrix > 0, axis=1).A1
    nnodes = adj_matrix.shape[0]

    neighbors, weights = calc_ppr_topk_parallel(
        adj_matrix.indptr, adj_matrix.indices, out_degree, numba.float32(alpha), numba.float32(epsilon), nodes, topk
    )

    return construct_sparse(neighbors, weights, (len(nodes), nnodes))


def construct_sparse(neighbors, weights, shape):
    i = np.repeat(np.arange(len(neighbors)), np.fromiter(map(len, neighbors), dtype=np.int))
    j = np.concatenate(neighbors)
    return sp.coo_matrix((np.concatenate(weights), (i, j)), shape)


def topk_ppr_matrix(adj_matrix, alpha, eps, idx, topk, normalization="row"):
    """Create a sparse matrix where each node has up to the topk PPR neighbors and their weights."""

    topk_matrix = ppr_topk(adj_matrix, alpha, eps, idx, topk).tocsr()

    if normalization == "sym":
        # Assume undirected (symmetric) adjacency matrix
        deg = adj_matrix.sum(1).A1
        deg_sqrt = np.sqrt(np.maximum(deg, 1e-12))
        deg_inv_sqrt = 1.0 / deg_sqrt

        row, col = topk_matrix.nonzero()
        topk_matrix.data = deg_sqrt[idx[row]] * topk_matrix.data * deg_inv_sqrt[col]
    elif normalization == "col":
        # Assume undirected (symmetric) adjacency matrix
        deg = adj_matrix.sum(1).A1
        deg_inv = 1.0 / np.maximum(deg, 1e-12)

        row, col = topk_matrix.nonzero()
        topk_matrix.data = deg[idx[row]] * topk_matrix.data * deg_inv[col]
    elif normalization == "row":
        pass
    else:
        raise ValueError(f"Unknown PPR normalization: {normalization}")

    return topk_matrix


def build_topk_ppr_matrix_from_data(edge_index, *args, **kwargs):
    if isinstance(edge_index, torch.Tensor) or isinstance(edge_index, tuple):
        row, col = edge_index
        row, col = row.numpy(), col.numpy()
        num_node = int(max(row.max(), col.max())) + 1

        val = np.ones(row.shape[0])
        adj_matrix = sp.csr_matrix((val, (row, col)), shape=(num_node, num_node))
    else:
        adj_matrix = edge_index
    return topk_ppr_matrix(adj_matrix, *args, **kwargs)



import importlib
import torch
import inspect

from cogdl.data.dataset import Dataset
from .customized_data import NodeDataset, GraphDataset, generate_random_graph


def register_dataset(name):
    """
    New dataset types can be added to cogdl with the :func:`register_dataset`
    function decorator.

    For example::

        @register_dataset('my_dataset')
        class MyDataset():
            (...)

    Args:
        name (str): the name of the dataset
    """

    def register_dataset_cls(cls):
        print("The `register_dataset` API is deprecated!")
        return cls

    return register_dataset_cls


def try_adding_dataset_args(dataset, parser):
    if dataset in SUPPORTED_DATASETS:
        path = ".".join(SUPPORTED_DATASETS[dataset].split(".")[:-1])
        module = importlib.import_module(path)
        class_name = SUPPORTED_DATASETS[dataset].split(".")[-1]
        dataset_class = getattr(module, class_name)
        if hasattr(dataset_class, "add_args"):
            dataset_class.add_args(parser)


def build_dataset_from_name(dataset, split=0):
    if isinstance(dataset, list):
        dataset = dataset[0]
    if isinstance(split, list):
        split = split[0]
    if dataset in SUPPORTED_DATASETS:
        path = ".".join(SUPPORTED_DATASETS[dataset].split(".")[:-1])
        module = importlib.import_module(path)
    else:
        dataset = build_dataset_from_path(dataset)
        if dataset is not None:
            return dataset
        raise NotImplementedError(f"Failed to import {dataset} dataset.")
    class_name = SUPPORTED_DATASETS[dataset].split(".")[-1]
    dataset_class = getattr(module, class_name)
    for key in inspect.signature(dataset_class.__init__).parameters.keys():
        if key == "split":
            return dataset_class(split=split)

    return dataset_class()


def build_dataset(args):
    if not hasattr(args, "split"):
        args.split = 0
    dataset = build_dataset_from_name(args.dataset, args.split)

    if hasattr(dataset, "num_classes") and dataset.num_classes > 0:
        args.num_classes = dataset.num_classes
    if hasattr(dataset, "num_features") and dataset.num_features > 0:
        args.num_features = dataset.num_features

    return dataset


def build_dataset_from_path(data_path, dataset=None):
    if dataset is not None and dataset in SUPPORTED_DATASETS:
        path = ".".join(SUPPORTED_DATASETS[dataset].split(".")[:-1])
        module = importlib.import_module(path)
        class_name = SUPPORTED_DATASETS[dataset].split(".")[-1]
        dataset = getattr(module, class_name)(data_path=data_path)
        return dataset

    if dataset is None:
        try:
            return torch.load(data_path)
        except Exception as e:
            print(e)
            exit(0)
    raise ValueError("You are expected to specify `dataset` and `data_path`")


SUPPORTED_DATASETS = {
    "kdd_icdm": "cogdl.datasets.gcc_data.KDD_ICDM_GCCDataset",
    "sigir_cikm": "cogdl.datasets.gcc_data.SIGIR_CIKM_GCCDataset",
    "sigmod_icde": "cogdl.datasets.gcc_data.SIGMOD_ICDE_GCCDataset",
    "usa-airport": "cogdl.datasets.gcc_data.USAAirportDataset",
    "ogbn-arxiv": "cogdl.datasets.ogb.OGBArxivDataset",
    "ogbn-products": "cogdl.datasets.ogb.OGBProductsDataset",
    "ogbn-proteins": "cogdl.datasets.ogb.OGBProteinsDataset",
    "ogbn-papers100M": "cogdl.datasets.ogb.OGBPapers100MDataset",
    "ogbg-molbace": "cogdl.datasets.ogb.OGBMolbaceDataset",
    "ogbg-molhiv": "cogdl.datasets.ogb.OGBMolhivDataset",
    "ogbg-molpcba": "cogdl.datasets.ogb.OGBMolpcbaDataset",
    "ogbg-ppa": "cogdl.datasets.ogb.OGBPpaDataset",
    "ogbg-code": "cogdl.datasets.ogb.OGBCodeDataset",
    "amazon": "cogdl.datasets.gatne.AmazonDataset",
    "twitter": "cogdl.datasets.gatne.TwitterDataset",
    "youtube": "cogdl.datasets.gatne.YouTubeDataset",
    "gtn-acm": "cogdl.datasets.gtn_data.ACM_GTNDataset",
    "gtn-dblp": "cogdl.datasets.gtn_data.DBLP_GTNDataset",
    "gtn-imdb": "cogdl.datasets.gtn_data.IMDB_GTNDataset",
    "fb13": "cogdl.datasets.kg_data.FB13Datset",
    "fb15k": "cogdl.datasets.kg_data.FB15kDatset",
    "fb15k237": "cogdl.datasets.kg_data.FB15k237Datset",
    "wn18": "cogdl.datasets.kg_data.WN18Datset",
    "wn18rr": "cogdl.datasets.kg_data.WN18RRDataset",
    "fb13s": "cogdl.datasets.kg_data.FB13SDatset",
    "cora": "cogdl.datasets.planetoid_data.CoraDataset",
    "citeseer": "cogdl.datasets.planetoid_data.CiteSeerDataset",
    "pubmed": "cogdl.datasets.planetoid_data.PubMedDataset",
    "chameleon": "cogdl.datasets.geom_data.ChameleonDataset",
    "cornell": "cogdl.datasets.geom_data.CornellDataset",
    "film": "cogdl.datasets.geom_data.FilmDataset",
    "squirrel": "cogdl.datasets.geom_data.SquirrelDataset",
    "texas": "cogdl.datasets.geom_data.TexasDataset",
    "wisconsin": "cogdl.datasets.geom_data.WisconsinDataset",
    "blogcatalog": "cogdl.datasets.matlab_matrix.BlogcatalogDataset",
    "flickr-ne": "cogdl.datasets.matlab_matrix.FlickrDataset",
    "dblp-ne": "cogdl.datasets.matlab_matrix.DblpNEDataset",
    "youtube-ne": "cogdl.datasets.matlab_matrix.YoutubeNEDataset",
    "wikipedia": "cogdl.datasets.matlab_matrix.WikipediaDataset",
    "ppi-ne": "cogdl.datasets.matlab_matrix.PPIDataset",
    "han-acm": "cogdl.datasets.han_data.ACM_HANDataset",
    "han-dblp": "cogdl.datasets.han_data.DBLP_HANDataset",
    "han-imdb": "cogdl.datasets.han_data.IMDB_HANDataset",
    "mutag": "cogdl.datasets.tu_data.MUTAGDataset",
    "imdb-b": "cogdl.datasets.tu_data.ImdbBinaryDataset",
    "imdb-m": "cogdl.datasets.tu_data.ImdbMultiDataset",
    "collab": "cogdl.datasets.tu_data.CollabDataset",
    "proteins": "cogdl.datasets.tu_data.ProteinsDataset",
    "reddit-b": "cogdl.datasets.tu_data.RedditBinary",
    "reddit-multi-5k": "cogdl.datasets.tu_data.RedditMulti5K",
    "reddit-multi-12k": "cogdl.datasets.tu_data.RedditMulti12K",
    "ptc-mr": "cogdl.datasets.tu_data.PTCMRDataset",
    "nci1": "cogdl.datasets.tu_data.NCI1Dataset",
    "nci109": "cogdl.datasets.tu_data.NCI109Dataset",
    "enzymes": "cogdl.datasets.tu_data.ENZYMES",
    "yelp": "cogdl.datasets.saint_data.YelpDataset",
    "amazon-s": "cogdl.datasets.saint_data.AmazonDataset",
    "flickr": "cogdl.datasets.saint_data.FlickrDataset",
    "reddit": "cogdl.datasets.saint_data.RedditDataset",
    "ppi": "cogdl.datasets.saint_data.PPIDataset",
    "ppi-large": "cogdl.datasets.saint_data.PPILargeDataset",
    "l0fos": "cogdl.datasets.oagbert_data.l0fos",
    "aff30": "cogdl.datasets.oagbert_data.aff30",
    "arxivvenue": "cogdl.datasets.oagbert_data.arxivvenue",
    "yelp2018": "cogdl.datasets.rec_data.Yelp2018Dataset",
    "ali": "cogdl.datasets.rec_data.AliDataset",
    "amazon-rec": "cogdl.datasets.rec_data.AmazonRecDataset",
    "Github": "cogdl.datasets.rd2cd_data.Github",
    "Elliptic": "cogdl.datasets.rd2cd_data.Elliptic",
    "Film": "cogdl.datasets.rd2cd_data.Film",
    "Wiki": "cogdl.datasets.rd2cd_data.Wiki",
    "Clothing": "cogdl.datasets.rd2cd_data.Clothing",
    "Electronics": "cogdl.datasets.rd2cd_data.Electronics",
    "Dblp": "cogdl.datasets.rd2cd_data.Dblp",
    "Yelpchi": "cogdl.datasets.rd2cd_data.Yelpchi",
    "Alpha": "cogdl.datasets.rd2cd_data.Alpha",
    "Weibo": "cogdl.datasets.rd2cd_data.Weibo",
    "bgp": "cogdl.datasets.rd2cd_data.bgp",
    "ssn5": "cogdl.datasets.rd2cd_data.ssn5",
    "ssn7": "cogdl.datasets.rd2cd_data.ssn7",
    "Aids": "cogdl.datasets.rd2cd_data.Aids",
    "Nba": "cogdl.datasets.rd2cd_data.Nba",
    "Pokec_z": "cogdl.datasets.rd2cd_data.Pokec_z",
}



import os.path as osp

import numpy as np
import scipy.io as sio
import torch
from cogdl.data import Graph, Dataset
from cogdl.utils import download_url, untar


def sample_mask(idx, length):
    """Create mask."""
    mask = np.zeros(length)
    mask[idx] = 1
    return np.array(mask, dtype=np.bool)


class HANDataset(Dataset):
    r"""The network datasets "ACM", "DBLP" and "IMDB" from the
    `"Heterogeneous Graph Attention Network"
    <https://arxiv.org/abs/1903.07293>`_ paper.

    Args:
        root (string): Root directory where the dataset should be saved.
        name (string): The name of the dataset (:obj:`"han-acm"`,
            :obj:`"han-dblp"`, :obj:`"han-imdb"`).
    """

    def __init__(self, root, name):
        self.name = name
        self.url = f"https://github.com/cenyk1230/han-data/blob/master/{name}.zip?raw=true"
        super(HANDataset, self).__init__(root)
        self.data = torch.load(self.processed_paths[0])
        self.num_edge = len(self.data.adj)
        self.num_nodes = self.data.x.shape[0]

    @property
    def raw_file_names(self):
        names = ["data.mat"]
        return names

    @property
    def processed_file_names(self):
        return ["data.pt"]

    @property
    def num_classes(self):
        return torch.max(self.data.train_target).item() + 1

    def read_gtn_data(self, folder):
        data = sio.loadmat(osp.join(folder, "data.mat"))
        if self.name == "han-acm" or self.name == "han-imdb":
            truelabels, truefeatures = data["label"], data["feature"].astype(float)
        elif self.name == "han-dblp":
            truelabels, truefeatures = data["label"], data["features"].astype(float)
        num_nodes = truefeatures.shape[0]
        if self.name == "han-acm":
            rownetworks = [data["PAP"] - np.eye(num_nodes), data["PLP"] - np.eye(num_nodes)]
        elif self.name == "han-dblp":
            rownetworks = [
                data["net_APA"] - np.eye(num_nodes),
                data["net_APCPA"] - np.eye(num_nodes),
                data["net_APTPA"] - np.eye(num_nodes),
            ]
        elif self.name == "han-imdb":
            rownetworks = [
                data["MAM"] - np.eye(num_nodes),
                data["MDM"] - np.eye(num_nodes),
                data["MYM"] - np.eye(num_nodes),
            ]

        y = truelabels
        train_idx = data["train_idx"]
        val_idx = data["val_idx"]
        test_idx = data["test_idx"]

        train_mask = sample_mask(train_idx, y.shape[0])
        val_mask = sample_mask(val_idx, y.shape[0])
        test_mask = sample_mask(test_idx, y.shape[0])

        y_train = np.argmax(y[train_mask, :], axis=1)
        y_val = np.argmax(y[val_mask, :], axis=1)
        y_test = np.argmax(y[test_mask, :], axis=1)

        data = Graph()
        A = []
        for i, edge in enumerate(rownetworks):
            edge_tmp = torch.from_numpy(np.vstack((edge.nonzero()[0], edge.nonzero()[1]))).type(torch.LongTensor)
            value_tmp = torch.ones(edge_tmp.shape[1]).type(torch.FloatTensor)
            A.append((edge_tmp, value_tmp))
        edge_tmp = torch.stack((torch.arange(0, num_nodes), torch.arange(0, num_nodes))).type(torch.LongTensor)
        value_tmp = torch.ones(num_nodes).type(torch.FloatTensor)
        A.append((edge_tmp, value_tmp))
        data.adj = A

        data.x = torch.from_numpy(truefeatures).type(torch.FloatTensor)

        data.train_node = torch.from_numpy(train_idx[0]).type(torch.LongTensor)
        data.train_target = torch.from_numpy(y_train).type(torch.LongTensor)
        data.valid_node = torch.from_numpy(val_idx[0]).type(torch.LongTensor)
        data.valid_target = torch.from_numpy(y_val).type(torch.LongTensor)
        data.test_node = torch.from_numpy(test_idx[0]).type(torch.LongTensor)
        data.test_target = torch.from_numpy(y_test).type(torch.LongTensor)

        y = np.zeros((num_nodes), dtype=int)
        x_index = torch.cat((data.train_node, data.valid_node, data.test_node))
        y_index = torch.cat((data.train_target, data.valid_target, data.test_target))
        y[x_index.numpy()] = y_index.numpy()
        data.y = torch.from_numpy(y)

        self.data = data

    def get(self, idx):
        assert idx == 0
        return self.data

    def apply_to_device(self, device):
        self.data.x = self.data.x.to(device)
        self.data.y = self.data.y.to(device)

        self.data.train_node = self.data.train_node.to(device)
        self.data.valid_node = self.data.valid_node.to(device)
        self.data.test_node = self.data.test_node.to(device)

        self.data.train_target = self.data.train_target.to(device)
        self.data.valid_target = self.data.valid_target.to(device)
        self.data.test_target = self.data.test_target.to(device)

        new_adj = []
        for (t1, t2) in self.data.adj:
            new_adj.append((t1.to(device), t2.to(device)))
        self.data.adj = new_adj

    def download(self):
        download_url(self.url, self.raw_dir, name=self.name + ".zip")
        untar(self.raw_dir, self.name + ".zip")

    def process(self):
        self.read_gtn_data(self.raw_dir)
        torch.save(self.data, self.processed_paths[0])

    def __repr__(self):
        return "{}()".format(self.name)


class ACM_HANDataset(HANDataset):
    def __init__(self, data_path="data"):
        dataset = "han-acm"
        path = osp.join(data_path, dataset)
        super(ACM_HANDataset, self).__init__(path, dataset)


class DBLP_HANDataset(HANDataset):
    def __init__(self, data_path="data"):
        dataset = "han-dblp"
        path = osp.join(data_path, dataset)
        super(DBLP_HANDataset, self).__init__(path, dataset)


class IMDB_HANDataset(HANDataset):
    def __init__(self, data_path="data"):
        dataset = "han-imdb"
        path = osp.join(data_path, dataset)
        super(IMDB_HANDataset, self).__init__(path, dataset)



import tvm
from tvm import te
from tvm.topi.util import get_const_tuple


def vanilla_sddmm(SrcFeat,
                  DstFeat,
                  Adj_row_indices,
                  Adj_col_indices,
                  num_feat_partitions=1):
    # TODO：apply parallelization in cpu schedule
    # TODO: support tuning both block number and thread number in cuda schedule
    """Compute sampled dense dense matrix multiplication of SrcFeat and DstFeat with Adj matrix as mask.

    Parameters
    ----------
    SrcFeat : tvm.te.Tensor
        2-D with shape [num_rows, feat_len]

    DstFeat : tvm.te.Tensor
        2-D with shape [num_cols, feat_len]

    Adj_row_indices : tvm.te.Tensor
        1-D with shape [nnz] (COO)

    Adj_col_indices : tvm.te.Tensor
        1-D with shape [nnz] (COO)

    num_feat_partitions : int
        Doing feature dimension tiling

    Returns
    -------
    Out : tvm.te.Tensor
        1-D with shape [nnz] (COO)
    """
    feat_len = get_const_tuple(SrcFeat.shape)[1]
    assert get_const_tuple(DstFeat.shape)[1] == feat_len, "dimension mismatch"
    num_edges = get_const_tuple(Adj_row_indices.shape)[0]
    assert get_const_tuple(Adj_col_indices.shape)[0] == num_edges, "dimension mismatch"
    oshape = (num_edges,)

    k = te.reduce_axis((0, feat_len))

    if num_feat_partitions == 1:
        def edgefunc(eid):  # eid: edge id
            return te.sum(SrcFeat[Adj_col_indices[eid], k] * DstFeat[Adj_row_indices[eid], k], axis=k)
    else:
        feat_len_per_partition = feat_len // num_feat_partitions  # we assume feat_len % num_feat_partitions = 0
        num_rows = get_const_tuple(SrcFeat.shape)[0]
        num_cols = get_const_tuple(DstFeat.shape)[0]
        ReshapedSrcFeat = te.compute((num_feat_partitions, num_rows, feat_len_per_partition), \
                                      lambda fo, nn, fi: SrcFeat[nn, fo*feat_len_per_partition + fi], \
                                      name='ReshapedSrcFeat')
        ReshapedDstFeat = te.compute((num_feat_partitions, num_cols, feat_len_per_partition), \
                                      lambda fo, nn, fi: DstFeat[nn, fo*feat_len_per_partition + fi], \
                                      name='ReshapedDstFeat')
        def edgefunc(eid):  # eid: edge id
            return te.sum(ReshapedSrcFeat[k // feat_len_per_partition, Adj_col_indices[eid], k % feat_len_per_partition] \
                          * ReshapedDstFeat[k // feat_len_per_partition, Adj_row_indices[eid], k % feat_len_per_partition], axis=k)

    Out = te.compute(oshape, edgefunc, name='vanilla_sddmm')
    return Out


def schedule_vanilla_sddmm_x86(Out, num_feat_partitions=1):
    s = te.create_schedule([Out.op])
    if num_feat_partitions != 1:
        edge_iter_axis = Out.op.axis[0]
        feat_reduce_axis = Out.op.reduce_axis[0]
        fo, fi = s[Out.op].split(feat_reduce_axis, nparts=num_feat_partitions)
        s[Out.op].reorder(fo, edge_iter_axis, fi)
        # TODO: parallelize ReshapedSrcFeat and ReshapedDstFeat
    return s


def schedule_vanilla_sddmm_cuda_tree_reduce(Out, num_feat_partitions=1, num_cuda_blocks=8192):
    s = te.create_schedule([Out.op])
    assert num_feat_partitions == 1, "cuda schedule for sddmm does not support feat dimension tiling, " \
                                     "which requires cross-cuda-block reduction and atomic operations."
    num_edges = Out.shape[0].value
    assert num_cuda_blocks <= num_edges, "num_cuda_blocks must be smaller than num_edges, " \
                                         "which is {}.".format(num_edges)
    edge_iter_axis = Out.op.axis[0]
    block_idx, _ = s[Out.op].split(edge_iter_axis, nparts=num_cuda_blocks)
    s[Out.op].bind(block_idx, te.thread_axis("blockIdx.x"))
    # Pay attention: here is doing tree reduce
    s[Out.op].bind(Out.op.reduce_axis[0], te.thread_axis("threadIdx.x"))
    return s


def schedule_vanilla_sddmm_cuda_single_thread_reduce(Out, num_feat_partitions=1, num_threads_per_cuda_block=32):
    s = te.create_schedule([Out.op])
    assert num_feat_partitions == 1, "cuda schedule for sddmm does not support feat dimension tiling, " \
                                     "which requires cross-cuda-block reduction and atomic operations."
    edge_iter_axis = Out.op.axis[0]
    block_idx, thread_idx = s[Out.op].split(edge_iter_axis, factor=num_threads_per_cuda_block)
    s[Out.op].bind(block_idx, te.thread_axis("blockIdx.x"))
    s[Out.op].bind(thread_idx, te.thread_axis("threadIdx.x"))
    return s



import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from cogdl.utils import (
    EdgeSoftmax,
    MultiHeadSpMM,
    get_activation,
    get_norm_layer,
    check_fused_gat,
    fused_gat_op,
)


class GATLayer(nn.Module):
    """
    Sparse version GAT layer, similar to https://arxiv.org/abs/1710.10903
    """

    def __init__(
        self, in_feats, out_feats, nhead=1, alpha=0.2, attn_drop=0.5, activation=None, residual=False, norm=None
    ):
        super(GATLayer, self).__init__()
        self.in_features = in_feats
        self.out_features = out_feats
        self.alpha = alpha
        self.nhead = nhead

        self.W = nn.Parameter(torch.FloatTensor(in_feats, out_feats * nhead))

        self.a_l = nn.Parameter(torch.zeros(size=(1, nhead, out_feats)))
        self.a_r = nn.Parameter(torch.zeros(size=(1, nhead, out_feats)))

        self.edge_softmax = EdgeSoftmax()
        self.mhspmm = MultiHeadSpMM()

        self.dropout = nn.Dropout(attn_drop)
        self.leakyrelu = nn.LeakyReLU(self.alpha)
        self.act = None if activation is None else get_activation(activation)
        self.norm = None if norm is None else get_norm_layer(norm, out_feats * nhead)

        if residual:
            self.residual = nn.Linear(in_feats, out_feats * nhead)
        else:
            self.register_buffer("residual", None)
        self.reset_parameters()

    def reset_parameters(self):
        def reset(tensor):
            stdv = math.sqrt(6.0 / (tensor.size(-2) + tensor.size(-1)))
            tensor.data.uniform_(-stdv, stdv)

        reset(self.a_l)
        reset(self.a_r)
        reset(self.W)

    def forward(self, graph, x):
        h = torch.matmul(x, self.W).view(-1, self.nhead, self.out_features)
        h[torch.isnan(h)] = 0.0

        row, col = graph.edge_index
        # Self-attention on the nodes - Shared attention mechanism
        h_l = (self.a_l * h).sum(dim=-1)
        h_r = (self.a_r * h).sum(dim=-1)

        if self.dropout.p == 0.0 and graph.is_symmetric() and check_fused_gat():
            out = fused_gat_op(h_l, h_r, graph, self.alpha, h)
            out = out.view(out.shape[0], -1)
        else:
            # edge_attention: E * H
            edge_attention = self.leakyrelu(h_l[row] + h_r[col])
            edge_attention = self.edge_softmax(graph, edge_attention)
            edge_attention = self.dropout(edge_attention)

            out = self.mhspmm(graph, edge_attention, h)

        if self.residual:
            res = self.residual(x)
            out += res
        if self.norm is not None:
            out = self.norm(out)
        if self.act is not None:
            out = self.act(out)
        return out

    def __repr__(self):
        return self.__class__.__name__ + " (" + str(self.in_features) + " -> " + str(self.out_features) + ")"



import copy
import warnings
from typing import Optional
import numpy as np
from tqdm import tqdm
import os

import torch
import torch.distributed as dist
from torch.nn.parallel import DistributedDataParallel
import torch.multiprocessing as mp

from cogdl.wrappers.data_wrapper.base_data_wrapper import DataWrapper
from cogdl.wrappers.model_wrapper.base_model_wrapper import ModelWrapper, EmbeddingModelWrapper
from cogdl.trainer.trainer_utils import evaluation_comp, load_model, save_model, ddp_end, ddp_after_epoch, Printer
from cogdl.trainer.embed_trainer import EmbeddingTrainer
from cogdl.trainer.controller import DataController
from cogdl.loggers import build_logger
from cogdl.data import Graph


def move_to_device(batch, device):
    if isinstance(batch, list) or isinstance(batch, tuple):
        if isinstance(batch, tuple):
            batch = list(batch)
        for i, x in enumerate(batch):
            if torch.is_tensor(x):
                batch[i] = x.to(device)
            elif isinstance(x, Graph):
                x.to(device)
    elif torch.is_tensor(batch) or isinstance(batch, Graph):
        batch = batch.to(device)
    elif hasattr(batch, "apply_to_device"):
        batch.apply_to_device(device)
    return batch


def clip_grad_norm(params, max_norm):
    """Clips gradient norm."""
    if max_norm > 0:
        return torch.nn.utils.clip_grad_norm_(params, max_norm)
    else:
        return torch.sqrt(sum(p.grad.data.norm() ** 2 for p in params if p.grad is not None))


class Trainer(object):
    def __init__(
        self,
        epochs: int,
        max_epoch: int = None,
        nstage: int = 1,
        cpu: bool = False,
        checkpoint_path: str = "./checkpoints/model.pt",
        resume_training: str = False,
        device_ids: Optional[list] = None,
        distributed_training: bool = False,
        distributed_inference: bool = False,
        master_addr: str = "localhost",
        master_port: int = 10086,
        early_stopping: bool = True,
        patience: int = 100,
        eval_step: int = 1,
        save_emb_path: Optional[str] = None,
        load_emb_path: Optional[str] = None,
        cpu_inference: bool = False,
        progress_bar: str = "epoch",
        clip_grad_norm: float = 5.0,
        logger: str = None,
        log_path: str = "./runs",
        project: str = "cogdl-exp",
        no_test: bool = False,
        actnn: bool = False,
        rp_ratio: int = 1,
    ):
        self.epochs = epochs
        self.nstage = nstage
        self.patience = patience
        self.early_stopping = early_stopping
        self.eval_step = eval_step
        self.monitor = None
        self.evaluation_metric = None
        self.progress_bar = progress_bar

        if max_epoch is not None:
            warnings.warn("The max_epoch is deprecated and will be removed in the future, please use epochs instead!")
            self.epochs = max_epoch

        self.cpu = cpu
        self.devices, self.world_size = self.set_device(device_ids)
        self.checkpoint_path = checkpoint_path
        self.resume_training = resume_training

        self.distributed_training = distributed_training
        self.distributed_inference = distributed_inference

        self.master_addr = master_addr
        self.master_port = master_port

        self.cpu_inference = cpu_inference

        self.no_test = no_test

        self.on_train_batch_transform = None
        self.on_eval_batch_transform = None
        self.clip_grad_norm = clip_grad_norm

        self.save_emb_path = save_emb_path
        self.load_emb_path = load_emb_path

        self.data_controller = DataController(world_size=self.world_size, distributed=self.distributed_training)

        self.logger = build_logger(logger, log_path, project)

        self.after_epoch_hooks = []
        self.pre_epoch_hooks = []
        self.training_end_hooks = []

        if distributed_training:
            self.register_training_end_hook(ddp_end)
            self.register_out_epoch_hook(ddp_after_epoch)

        self.eval_data_back_to_cpu = False

        if actnn:
            try:
                import actnn
                from actnn.conf import config

                actnn.set_optimization_level("L3")
                if rp_ratio > 1:
                    config.group_size = 64
            except Exception:
                pass

    def register_in_epoch_hook(self, hook):
        self.pre_epoch_hooks.append(hook)

    def register_out_epoch_hook(self, hook):
        self.after_epoch_hooks.append(hook)

    def register_training_end_hook(self, hook):
        self.training_end_hooks.append(hook)

    def set_device(self, device_ids: Optional[list]):
        """
        Return: devices, world_size
        """
        if device_ids is None or self.cpu:
            return [torch.device("cpu")], 0

        if isinstance(device_ids, int) and device_ids > 0:
            device_ids = [device_ids]
        elif isinstance(device_ids, list):
            pass
        else:
            raise ValueError("`device_id` has to be list of integers")
        if len(device_ids) == 0:
            return torch.device("cpu"), 0
        else:
            return [i for i in device_ids], len(device_ids)

    def run(self, model_w: ModelWrapper, dataset_w: DataWrapper):
        # for network/graph embedding models
        if isinstance(model_w, EmbeddingModelWrapper):
            return EmbeddingTrainer(self.save_emb_path, self.load_emb_path).run(model_w, dataset_w)

        print("Model Parameters:", sum(p.numel() for p in model_w.parameters()))

        # for deep learning models
        # set default loss_fn and evaluator for model_wrapper
        # mainly for in-cogdl setting
        model_w.default_loss_fn = dataset_w.get_default_loss_fn()
        model_w.default_evaluator = dataset_w.get_default_evaluator()
        model_w.set_evaluation_metric()

        if self.resume_training:
            model_w = load_model(model_w, self.checkpoint_path).to(self.devices[0])

        if self.distributed_training:
            torch.multiprocessing.set_sharing_strategy("file_system")
            self.dist_train(model_w, dataset_w)
        else:
            self.train(self.devices[0], model_w, dataset_w)
        best_model_w = load_model(model_w, self.checkpoint_path).to(self.devices[0])

        if self.no_test:
            return best_model_w.model

        final_test = self.evaluate(best_model_w, dataset_w)
        return final_test

    def evaluate(self, model_w: ModelWrapper, dataset_w: DataWrapper, cpu=False):
        if cpu:
            self.devices = [torch.device("cpu")]

        # disable `distributed` to inference once only
        self.distributed_training = False
        dataset_w.prepare_test_data()
        final_val = self.validate(model_w, dataset_w, self.devices[0])
        final_test = self.test(model_w, dataset_w, self.devices[0])

        if final_val is not None and "val_metric" in final_val:
            final_val[f"val_{self.evaluation_metric}"] = final_val["val_metric"]
            final_val.pop("val_metric")
            if "val_loss" in final_val:
                final_val.pop("val_loss")

        if final_test is not None and "test_metric" in final_test:
            final_test[f"test_{self.evaluation_metric}"] = final_test["test_metric"]
            final_test.pop("test_metric")
            if "test_loss" in final_test:
                final_test.pop("test_loss")

        self.logger.note(final_test)
        if final_val is not None:
            final_test.update(final_val)
        print(final_test)
        return final_test

    def dist_train(self, model_w: ModelWrapper, dataset_w: DataWrapper):
        mp.set_start_method("spawn", force=True)

        device_count = torch.cuda.device_count()
        if device_count < self.world_size:
            size = device_count
            print(f"Available device count ({device_count}) is less than world size ({self.world_size})")
        else:
            size = self.world_size

        print(f"Let's using {size} GPUs.")

        processes = []
        for rank in range(size):
            p = mp.Process(target=self.train, args=(rank, model_w, dataset_w))

            p.start()
            print(f"Process [{rank}] starts!")
            processes.append(p)

        for p in processes:
            p.join()

    def build_optimizer(self, model_w):
        opt_wrap = model_w.setup_optimizer()
        if isinstance(opt_wrap, list) or isinstance(opt_wrap, tuple):
            assert len(opt_wrap) == 2
            optimizers, lr_schedulars = opt_wrap
        else:
            optimizers = opt_wrap
            lr_schedulars = None

        if not isinstance(optimizers, list):
            optimizers = [optimizers]
        if lr_schedulars and not isinstance(lr_schedulars, list):
            lr_schedulars = [lr_schedulars]
        return optimizers, lr_schedulars

    def initialize(self, model_w, rank=0, master_addr: str = "localhost", master_port: int = 10008):
        if self.distributed_training:
            os.environ["MASTER_ADDR"] = master_addr
            os.environ["MASTER_PORT"] = str(master_port)
            dist.init_process_group("nccl", rank=rank, world_size=self.world_size)
            model_w = copy.deepcopy(model_w).to(rank)
            model_w = DistributedDataParallel(model_w, device_ids=[rank])

            module = model_w.module
            model_w, model_ddp = module, model_w
            return model_w, model_ddp
        else:
            return model_w.to(rank), None

    def train(self, rank, model_w, dataset_w):
        model_w, _ = self.initialize(model_w, rank=rank, master_addr=self.master_addr, master_port=self.master_port)
        self.data_controller.prepare_data_wrapper(dataset_w, rank)
        self.eval_data_back_to_cpu = dataset_w.data_back_to_cpu

        optimizers, lr_schedulars = self.build_optimizer(model_w)
        if optimizers[0] is None:
            return

        est = model_w.set_early_stopping()
        if isinstance(est, str):
            est_monitor = est
            best_index, compare_fn = evaluation_comp(est_monitor)
        else:
            assert len(est) == 2
            est_monitor, est_compare = est
            best_index, compare_fn = evaluation_comp(est_monitor, est_compare)
        self.monitor = est_monitor
        self.evaluation_metric = model_w.evaluation_metric

        best_model_w = None

        patience = 0
        best_epoch = 0
        for stage in range(self.nstage):
            with torch.no_grad():
                pre_stage_out = model_w.pre_stage(stage, dataset_w)
                dataset_w.pre_stage(stage, pre_stage_out)
                self.data_controller.training_proc_per_stage(dataset_w, rank)

            if self.progress_bar == "epoch":
                epoch_iter = tqdm(range(self.epochs))
                epoch_printer = Printer(epoch_iter.set_description, rank=rank, world_size=self.world_size)
            else:
                epoch_iter = range(self.epochs)
                epoch_printer = Printer(print, rank=rank, world_size=self.world_size)

            self.logger.start()
            print_str_dict = dict()
            for epoch in epoch_iter:
                for hook in self.pre_epoch_hooks:
                    hook(self)

                # inductive setting ..
                dataset_w.train()
                train_loader = dataset_w.on_train_wrapper()
                training_loss = self.training_step(model_w, train_loader, optimizers, lr_schedulars, rank)

                print_str_dict["Epoch"] = epoch
                print_str_dict["train_loss"] = training_loss

                val_loader = dataset_w.on_val_wrapper()
                if val_loader is not None and epoch % self.eval_step == 0:
                    # inductive setting ..
                    dataset_w.eval()
                    # do validation in inference device
                    val_result = self.validate(model_w, dataset_w, rank)
                    if val_result is not None:
                        monitoring = val_result[self.monitor]
                        if compare_fn(monitoring, best_index):
                            best_index = monitoring
                            best_epoch = epoch
                            patience = 0
                            best_model_w = copy.deepcopy(model_w)
                        else:
                            patience += 1
                            if self.early_stopping and patience >= self.patience:
                                break
                        print_str_dict[f"val_{self.evaluation_metric}"] = monitoring

                if self.distributed_training:
                    if rank == 0:
                        epoch_printer(print_str_dict)
                        self.logger.note(print_str_dict, epoch)
                else:
                    epoch_printer(print_str_dict)
                    self.logger.note(print_str_dict, epoch)

                for hook in self.after_epoch_hooks:
                    hook(self)

            with torch.no_grad():
                model_w.eval()
                post_stage_out = model_w.post_stage(stage, dataset_w)
                dataset_w.post_stage(stage, post_stage_out)

            if best_model_w is None:
                best_model_w = copy.deepcopy(model_w)

        if self.distributed_training:
            if rank == 0:
                save_model(best_model_w.to("cpu"), self.checkpoint_path, best_epoch)
                dist.barrier()
            else:
                dist.barrier()
        else:
            save_model(best_model_w.to("cpu"), self.checkpoint_path, best_epoch)

        for hook in self.training_end_hooks:
            hook(self)

    def validate(self, model_w: ModelWrapper, dataset_w: DataWrapper, device):
        # ------- distributed training ---------
        if self.distributed_training:
            return self.distributed_test(model_w, dataset_w.on_val_wrapper(), device, self.val_step)
        # ------- distributed training ---------

        model_w.eval()
        dataset_w.eval()
        if self.cpu_inference:
            model_w.to("cpu")
            _device = device
        else:
            _device = device

        val_loader = dataset_w.on_val_wrapper()
        result = self.val_step(model_w, val_loader, _device)

        model_w.to(device)
        return result

    def test(self, model_w: ModelWrapper, dataset_w: DataWrapper, device):
        # ------- distributed training ---------
        if self.distributed_training:
            return self.distributed_test(model_w, dataset_w.on_test_wrapper(), device, self.test_step)
        # ------- distributed training ---------

        model_w.eval()
        dataset_w.eval()
        if self.cpu_inference:
            model_w.to("cpu")
            _device = device
        else:
            _device = device

        test_loader = dataset_w.on_test_wrapper()
        result = self.test_step(model_w, test_loader, _device)

        model_w.to(device)
        return result

    def distributed_test(self, model_w: ModelWrapper, loader, rank, fn):
        model_w.eval()
        # if rank == 0:
        if dist.get_rank() == 0:
            if self.cpu_inference:
                model_w.to("cpu")
                _device = "cpu"
            else:
                _device = rank
            result = fn(model_w, loader, _device)
            model_w.to(rank)

            object_list = [result]
        else:
            object_list = [None]
        dist.broadcast_object_list(object_list, src=0)
        return object_list[0]

    def training_step(self, model_w, train_loader, optimizers, lr_schedulars, device):
        model_w.train()
        losses = []

        if self.progress_bar == "iteration":
            train_loader = tqdm(train_loader)

        for batch in train_loader:
            batch = move_to_device(batch, device)
            loss = model_w.on_train_step(batch)

            for optimizer in optimizers:
                optimizer.zero_grad()
            loss.backward()

            torch.nn.utils.clip_grad_norm_(model_w.parameters(), self.clip_grad_norm)

            for optimizer in optimizers:
                optimizer.step()

            losses.append(loss.item())
        if lr_schedulars is not None:
            for lr_schedular in lr_schedulars:
                lr_schedular.step()
        return np.mean(losses)

    @torch.no_grad()
    def val_step(self, model_w, val_loader, device):
        model_w.eval()
        if val_loader is None:
            return None
        for batch in val_loader:
            batch = move_to_device(batch, device)
            model_w.on_val_step(batch)
            if self.eval_data_back_to_cpu:
                move_to_device(batch, "cpu")
        return model_w.collect_notes()

    # @torch.no_grad()
    def test_step(self, model_w, test_loader, device):
        model_w.eval()
        if test_loader is None:
            return None
        for batch in test_loader:
            batch = move_to_device(batch, device)
            model_w.on_test_step(batch)
            if self.eval_data_back_to_cpu:
                move_to_device(batch, "cpu")
        return model_w.collect_notes()



from .cluster_dw import ClusterWrapper
from .graphsage_dw import GraphSAGEDataWrapper
from .m3s_dw import M3SDataWrapper
from .network_embedding_dw import NetworkEmbeddingDataWrapper
from .node_classification_dw import FullBatchNodeClfDataWrapper
from .pprgo_dw import PPRGoDataWrapper
from .sagn_dw import SAGNDataWrapper



import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F

from cogdl.layers import SELayer
from .. import BaseModel

from cogdl.layers import MLP, GATLayer, GINLayer
from cogdl.utils import batch_sum_pooling, batch_mean_pooling, batch_max_pooling
from cogdl.layers import Set2Set


class ApplyNodeFunc(nn.Module):
    """Update the node feature hv with MLP, BN and ReLU."""

    def __init__(self, mlp, use_selayer):
        super(ApplyNodeFunc, self).__init__()
        self.mlp = mlp
        self.bn = (
            SELayer(self.mlp.output_dim, int(np.sqrt(self.mlp.output_dim)))
            if use_selayer
            else nn.BatchNorm1d(self.mlp.output_dim)
        )

    def forward(self, h):
        h = self.mlp(h)
        h = self.bn(h)
        h = F.relu(h)
        return h


class GATModel(nn.Module):
    def __init__(self, in_feats, hidden_size, num_layers, nhead, dropout=0.0, attn_drop=0.0, alpha=0.2, residual=False):
        super(GATModel, self).__init__()
        assert hidden_size % nhead == 0
        self.layers = nn.ModuleList(
            [
                GATLayer(
                    in_feats=in_feats if i > 0 else hidden_size // nhead,
                    out_feats=hidden_size // nhead,
                    nhead=nhead,
                    attn_drop=0.0,
                    alpha=0.2,
                    residual=False,
                    activation=F.leaky_relu if i + 1 < num_layers else None,
                )
                for i in range(num_layers)
            ]
        )

    def forward(self, graph, x):
        for i, layer in enumerate(self.layers):
            x = layer(graph, x)
        return x


class GINModel(nn.Module):
    def __init__(
        self,
        num_layers,
        in_feats,
        hidden_dim,
        out_feats,
        num_mlp_layers,
        eps=0,
        pooling="sum",
        train_eps=False,
        dropout=0.5,
        final_dropout=0.2,
    ):
        super(GINModel, self).__init__()
        self.gin_layers = nn.ModuleList()
        self.batch_norm = nn.ModuleList()
        self.num_layers = num_layers
        for i in range(num_layers - 1):
            if i == 0:
                mlp = MLP(in_feats, hidden_dim, hidden_dim, num_mlp_layers, norm="batchnorm")
            else:
                mlp = MLP(hidden_dim, hidden_dim, hidden_dim, num_mlp_layers, norm="batchnorm")
            self.gin_layers.append(GINLayer(mlp, eps, train_eps))
            self.batch_norm.append(nn.BatchNorm1d(hidden_dim))

        self.linear_prediction = nn.ModuleList()
        for i in range(self.num_layers):
            if i == 0:
                self.linear_prediction.append(nn.Linear(in_feats, out_feats))
            else:
                self.linear_prediction.append(nn.Linear(hidden_dim, out_feats))
        self.dropout = nn.Dropout(dropout)

        if pooling == "sum":
            self.pool = batch_sum_pooling
        elif pooling == "mean":
            self.pool = batch_mean_pooling
        elif pooling == "max":
            self.pool = batch_max_pooling
        else:
            raise NotImplementedError
        self.final_drop = nn.Dropout(final_dropout)

    def forward(self, batch, n_feat):
        h = n_feat
        # device = h.device
        # batchsize = int(torch.max(batch.batch)) + 1

        layer_rep = [h]
        for i in range(self.num_layers - 1):
            h = self.gin_layers[i](batch, h)
            h = self.batch_norm[i](h)
            h = F.relu(h)
            layer_rep.append(h)

        score_over_layer = 0

        all_outputs = []
        for i, h in enumerate(layer_rep):
            pooled_h = self.pool(h, batch.batch)
            all_outputs.append(pooled_h)
            score_over_layer += self.final_drop(self.linear_prediction[i](pooled_h))

        return score_over_layer, all_outputs[1:]


class GCCModel(BaseModel):
    """
    MPNN from
    `Neural Message Passing for Quantum Chemistry <https://arxiv.org/abs/1704.01212>`__
    Parameters
    ----------
    node_input_dim : int
        Dimension of input node feature, default to be 15.
    edge_input_dim : int
        Dimension of input edge feature, default to be 15.
    output_dim : int
        Dimension of prediction, default to be 12.
    node_hidden_dim : int
        Dimension of node feature in hidden layers, default to be 64.
    edge_hidden_dim : int
        Dimension of edge feature in hidden layers, default to be 128.
    num_step_message_passing : int
        Number of message passing steps, default to be 6.
    num_step_set2set : int
        Number of set2set steps
    num_layer_set2set : int
        Number of set2set layers
    """

    @staticmethod
    def add_args(parser):
        parser.add_argument("--hidden-size", type=int, default=64)
        parser.add_argument("--positional-embedding-size", type=int, default=32)
        parser.add_argument("--degree-embedding-size", type=int, default=16)
        parser.add_argument("--max-node-freq", type=int, default=16)
        parser.add_argument("--max-edge-freq", type=int, default=16)
        parser.add_argument("--max-degree", type=int, default=512)
        parser.add_argument("--freq-embedding-size", type=int, default=16)
        parser.add_argument("--num-layers", type=int, default=2)
        parser.add_argument("--num-heads", type=int, default=2)
        parser.add_argument("--output-size", type=int, default=32)

    @classmethod
    def build_model_from_args(cls, args):
        return cls(
            positional_embedding_size=args.positional_embedding_size,
            max_node_freq=args.max_node_freq,
            max_edge_freq=args.max_edge_freq,
            max_degree=args.max_degree,
            num_layers=args.num_layers,
            num_heads=args.num_heads,
            degree_embedding_size=args.degree_embedding_size,
            node_hidden_dim=args.hidden_size,
            output_dim=args.output_size,
        )

    def __init__(
        self,
        positional_embedding_size=32,
        max_node_freq=8,
        max_edge_freq=8,
        max_degree=128,
        freq_embedding_size=32,
        degree_embedding_size=32,
        output_dim=32,
        node_hidden_dim=32,
        edge_hidden_dim=32,
        num_layers=6,
        num_heads=4,
        num_step_set2set=6,
        num_layer_set2set=3,
        norm=False,
        gnn_model="gin",
        degree_input=False,
    ):
        super(GCCModel, self).__init__()

        if degree_input:
            node_input_dim = positional_embedding_size + degree_embedding_size + 1
        else:
            node_input_dim = positional_embedding_size + 1
        # node_input_dim = (
        #     positional_embedding_size + freq_embedding_size + degree_embedding_size + 3
        # )
        # edge_input_dim = freq_embedding_size + 1
        if gnn_model == "gat":
            self.gnn = GATModel(
                in_feats=node_input_dim,
                hidden_size=node_hidden_dim,
                num_layers=num_layers,
                nhead=num_heads,
                dropout=0.0,
            )
        elif gnn_model == "gin":
            self.gnn = GINModel(
                num_layers=num_layers,
                num_mlp_layers=2,
                in_feats=node_input_dim,
                hidden_dim=node_hidden_dim,
                out_feats=output_dim,
                final_dropout=0.5,
                train_eps=False,
                pooling="sum",
                # neighbor_pooling_type="sum",
                # use_selayer=False,
            )
        self.gnn_model = gnn_model

        self.max_node_freq = max_node_freq
        self.max_edge_freq = max_edge_freq
        self.max_degree = max_degree
        self.degree_input = degree_input

        # self.node_freq_embedding = nn.Embedding(
        #     num_embeddings=max_node_freq + 1, embedding_dim=freq_embedding_size
        # )
        if degree_input:
            self.degree_embedding = nn.Embedding(num_embeddings=max_degree + 1, embedding_dim=degree_embedding_size)

        # self.edge_freq_embedding = nn.Embedding(
        #     num_embeddings=max_edge_freq + 1, embedding_dim=freq_embedding_size
        # )

        self.set2set = Set2Set(node_hidden_dim, num_step_set2set, num_layer_set2set)
        if gnn_model != "gin":
            self.lin_readout = nn.Sequential(
                nn.Linear(2 * node_hidden_dim, node_hidden_dim),
                nn.ReLU(),
                nn.Linear(node_hidden_dim, output_dim),
            )
        else:
            self.lin_readout = None
        self.norm = norm

    def forward(self, g, return_all_outputs=False):
        """Predict molecule labels
        Parameters
        ----------
        g : Graph
        n_feat : tensor of dtype float32 and shape (B1, D1)
            Node features. B1 for number of nodes and D1 for
            the node feature size.
        e_feat : tensor of dtype float32 and shape (B2, D2)
            Edge features. B2 for number of edges and D2 for
            the edge feature size.
        Returns
        -------
        res : Predicted labels
        """

        # nfreq = g.ndata["nfreq"]
        device = self.device
        pos_undirected = g.pos_undirected
        seed_emb = g.seed.unsqueeze(1).float()
        if not torch.is_tensor(seed_emb):
            seed_emb = torch.Tensor(seed_emb)

        if self.degree_input:
            degrees = g.degrees()
            if device != torch.device("cpu"):
                degrees = degrees.cuda(device)

            deg_emb = self.degree_embedding(degrees.clamp(0, self.max_degree))

            n_feat = torch.cat((pos_undirected, deg_emb, seed_emb), dim=-1)
        else:
            n_feat = torch.cat(
                (
                    pos_undirected,
                    # self.node_freq_embedding(nfreq.clamp(0, self.max_node_freq)),
                    # self.degree_embedding(degrees.clamp(0, self.max_degree)),
                    seed_emb,
                    # nfreq.unsqueeze(1).float() / self.max_node_freq,
                    # degrees.unsqueeze(1).float() / self.max_degree,
                ),
                dim=-1,
            )

        if self.gnn_model == "gin":
            x, all_outputs = self.gnn(g, n_feat)
        else:
            x, all_outputs = self.gnn(g, n_feat), None
            x = self.set2set(g, x)
            x = self.lin_readout(x)
        if self.norm:
            x = F.normalize(x, p=2, dim=-1, eps=1e-5)
        if return_all_outputs:
            return x, all_outputs
        else:
            return x


# --------------------------------------------------------------------
# --------------------------------------------------------------------


def warmup_linear(x, warmup=0.002):
    """Specifies a triangular learning rate schedule where peak is reached at `warmup`*`t_total`-th (as provided to BertAdam) training step.
    After `t_total`-th training step, learning rate is zero."""
    if x < warmup:
        return x / warmup
    return max((x - 1.0) / (warmup - 1.0), 0)



import torch
import torch.nn as nn

from .. import ModelWrapper
from cogdl.wrappers.tools.wrapper_utils import evaluate_node_embeddings_using_logreg


class MVGRLModelWrapper(ModelWrapper):
    def __init__(self, model, optimizer_cfg):
        super(MVGRLModelWrapper, self).__init__()
        self.model = model
        self.optimizer_cfg = optimizer_cfg
        self.loss_f = nn.BCEWithLogitsLoss()

    def train_step(self, subgraph):
        graph = subgraph
        logits = self.model(graph)
        labels = torch.zeros_like(logits)
        num_outs = logits.shape[1]
        labels[:, : num_outs // 2] = 1
        loss = self.loss_f(logits, labels)
        return loss

    def test_step(self, graph):
        with torch.no_grad():
            pred = self.model(graph)
        y = graph.y
        result = evaluate_node_embeddings_using_logreg(pred, y, graph.train_mask, graph.test_mask)
        self.note("test_acc", result)

    def setup_optimizer(self):
        cfg = self.optimizer_cfg
        return torch.optim.Adam(self.parameters(), lr=cfg["lr"], weight_decay=cfg["weight_decay"])



import torch
import torch.nn as nn
import torch.nn.functional as F

from cogdl.utils import row_normalization, spmm


class RGCNLayer(nn.Module):
    """
    Implementation of Relational-GCN in paper `"Modeling Relational Data with Graph Convolutional Networks"
    <https://arxiv.org/abs/1703.06103>`_

    Parameters
    ----------
    in_feats : int
        Size of each input embedding.
    out_feats : int
        Size of each output embedding.
    num_edge_type : int
        The number of edge type in knowledge graph.
    regularizer : str, optional
        Regularizer used to avoid overfitting, ``basis`` or ``bdd``, default : ``basis``.
    num_bases : int, optional
        The number of basis, only used when `regularizer` is `basis`, default : ``None``.
    self_loop : bool, optional
        Add self loop embedding if True, default : ``True``.
    dropout : float
    self_dropout : float, optional
        Dropout rate of self loop embedding, default : ``0.0``
    layer_norm : bool, optional
        Use layer normalization if True, default : ``True``
    bias : bool
    """

    def __init__(
        self,
        in_feats,
        out_feats,
        num_edge_types,
        regularizer="basis",
        num_bases=None,
        self_loop=True,
        dropout=0.0,
        self_dropout=0.0,
        layer_norm=True,
        bias=True,
    ):
        super(RGCNLayer, self).__init__()
        self.num_bases = num_bases
        self.regularizer = regularizer
        self.num_edge_types = num_edge_types
        self.in_feats = in_feats
        self.out_feats = out_feats
        self.self_loop = self_loop
        self.dropout = dropout
        self.self_dropout = self_dropout

        if self.num_bases is None or self.num_bases > num_edge_types or self.num_bases < 0:
            self.num_bases = num_edge_types

        if regularizer == "basis":
            self.weight = nn.Parameter(torch.Tensor(self.num_bases, in_feats, out_feats))
            if self.num_bases < num_edge_types:
                self.alpha = nn.Parameter(torch.Tensor(num_edge_types, self.num_bases))
            else:
                self.register_buffer("alpha", None)
        elif regularizer == "bdd":
            assert (in_feats % num_bases == 0) and (out_feats % num_bases == 0)
            self.block_in_feats = in_feats // num_bases
            self.block_out_feats = out_feats // num_bases
            self.weight = nn.Parameter(
                torch.Tensor(num_edge_types, self.num_bases, self.block_in_feats * self.block_out_feats)
            )
        else:
            raise NotImplementedError

        if bias is True:
            self.bias = nn.Parameter(torch.Tensor(out_feats))
        else:
            self.register_buffer("bias", None)

        if self_loop:
            self.weight_self_loop = nn.Parameter(torch.Tensor(in_feats, out_feats))
        else:
            self.register_buffer("weight_self_loop", None)

        if layer_norm:
            self.layer_norm = nn.LayerNorm(out_feats, elementwise_affine=True)
        else:
            self.register_buffer("layer_norm", None)
        self.reset_parameters()

    def reset_parameters(self):
        nn.init.xavier_uniform_(self.weight, gain=nn.init.calculate_gain("relu"))
        if self.alpha is not None:
            nn.init.xavier_uniform_(self.alpha, gain=nn.init.calculate_gain("relu"))
        if self.bias is not None:
            nn.init.zeros_(self.bias)
        if self.self_loop is not None:
            nn.init.xavier_uniform_(self.weight_self_loop, gain=nn.init.calculate_gain("relu"))

    def forward(self, graph, x):
        if self.regularizer == "basis":
            h_list = self.basis_forward(graph, x)
        else:
            h_list = self.bdd_forward(graph, x)

        h_result = sum(h_list)
        h_result = F.dropout(h_result, p=self.dropout, training=self.training)
        if self.layer_norm is not None:
            h_result = self.layer_norm(h_result)
        if self.bias is not None:
            h_result = h_result + self.bias
        if self.self_loop is not None:
            h_result += F.dropout(torch.matmul(x, self.weight_self_loop), p=self.self_dropout, training=self.training)
        return h_result

    def basis_forward(self, graph, x):
        edge_type = graph.edge_attr

        if self.num_bases < self.num_edge_types:
            weight = torch.matmul(self.alpha, self.weight.view(self.num_bases, -1))
            weight = weight.view(self.num_edge_types, self.in_feats, self.out_feats)
        else:
            weight = self.weight

        edge_index = torch.stack(graph.edge_index)
        edge_weight = graph.edge_weight

        graph.row_norm()
        h = torch.matmul(x, weight)  # (N, d1) by (r, d1, d2) -> (r, N, d2)

        h_list = []
        for edge_t in range(self.num_edge_types):
            g = graph.__class__()
            edge_mask = edge_type == edge_t

            if edge_mask.sum() == 0:
                h_list.append(0)
                continue

            g.edge_index = edge_index[:, edge_mask]

            g.edge_weight = edge_weight[edge_mask]
            g.padding_self_loops()

            temp = spmm(graph, h[edge_t])
            h_list.append(temp)
            return h_list

    def bdd_forward(self, graph, x):
        edge_type = graph.edge_attr
        edge_index = torch.stack(graph.edge_index)
        _x = x.view(-1, self.num_bases, self.block_in_feats)

        edge_weight = torch.ones(edge_type.shape).to(x.device)
        edge_weight = row_normalization(x.shape[0], edge_index, edge_weight)

        h_list = []
        for edge_t in range(self.num_edge_types):
            _weight = self.weight[edge_t].view(self.num_bases, self.block_in_feats, self.block_out_feats)
            edge_mask = edge_type == edge_t
            _edge_index_t = edge_index.t()[edge_mask].t()
            h_t = torch.einsum("abc,bcd->abd", _x, _weight).reshape(-1, self.out_feats)
            h_t = spmm(_edge_index_t, edge_weight[edge_mask], h_t)
            h_list.append(h_t)

        return h_list



import numpy as np

import torch
import torch.nn as nn
import torch.nn.functional as F


def cal_mrr(embedding, rel_embedding, edge_index, edge_type, scoring, protocol="raw", batch_size=1000, hits=[]):
    with torch.no_grad():
        if protocol == "raw":
            heads = edge_index[0]
            tails = edge_index[1]
            ranks_h = get_raw_rank(heads, tails, edge_type, embedding, rel_embedding, batch_size, scoring)
            ranks_t = get_raw_rank(tails, heads, edge_type, embedding, rel_embedding, batch_size, scoring)
            # ranks = torch.cat((ranks_h, ranks_t)) + 1
            ranks = np.concatenate((ranks_h, ranks_t)) + 1
        elif protocol == "filtered":
            raise NotImplementedError
        else:
            raise ValueError
        mrr = (1.0 / ranks).mean()
        hits_count = []
        # for hit in hits:
        # hits_count.append(torch.mean((ranks <= hit).float()).item())
        for hit in hits:
            hits_count.append(np.mean((ranks <= hit).astype(np.float)))
    # return mrr.item(), hits_count
    return mrr, hits_count


class DistMultLayer(nn.Module):
    def __init__(self):
        super(DistMultLayer, self).__init__()

    def forward(self, sub_emb, obj_emb, rel_emb):
        return torch.sum(sub_emb * obj_emb * rel_emb, dim=-1)

    def predict(self, sub_emb, obj_emb, rel_emb):
        return torch.matmul(sub_emb * rel_emb, obj_emb.t())


class ConvELayer(nn.Module):
    def __init__(self, dim, num_filter=20, kernel_size=7, k_w=10, dropout=0.3):
        super(ConvELayer, self).__init__()
        assert dim % k_w == 0
        self.k_w = k_w
        self.k_h = dim // k_w
        self.dim = dim
        self.bn0 = torch.nn.BatchNorm2d(1)
        self.bn1 = torch.nn.BatchNorm2d(num_filter)
        self.bn2 = torch.nn.BatchNorm1d(dim)

        self.hidden_drop = nn.Dropout(dropout)
        self.hidden_drop2 = nn.Dropout(dropout)
        self.feature_drop = nn.Dropout(dropout)
        self.conv = nn.Conv2d(
            1, out_channels=num_filter, kernel_size=(kernel_size, kernel_size), stride=1, padding=0, bias=True
        )

        flat_size_h = int(2 * self.k_w) - kernel_size + 1
        flat_size_w = self.k_h - kernel_size + 1
        self.flat_size = flat_size_h * flat_size_w * num_filter
        self.fc = nn.Linear(self.flat_size, dim)

        self.bias = nn.Parameter(torch.zeros(dim))

    def concat(self, ent, rel):
        ent = ent.view(-1, 1, self.dim)
        rel = rel.view(-1, 1, self.dim)
        ent_rel = torch.cat([ent, rel], dim=1)
        ent_rel = ent_rel.transpose(2, 1).reshape(-1, 1, 2 * self.k_w, self.k_h)
        return ent_rel

    def forward(self, sub_emb, obj_emb, rel_emb):
        h = self.concat(sub_emb, rel_emb)
        h = self.bn0(h)
        h = self.conv(h)
        h = F.relu(self.bn1(h))
        h = self.feature_drop(h)
        h = h.view(-1, self.flat_size)
        h = self.hidden_drop(self.fc(h))
        h = F.relu(self.bn2(self.hidden_drop2(h)))
        x = torch.sum(h * obj_emb + self.bias, dim=-1)
        return x

    def predict(self, sub_emb, obj_emb, rel_emb):
        h = self.concat(sub_emb, rel_emb)
        h = self.bn0(h)
        h = self.conv(h)
        h = F.relu(self.bn1(h))
        h = h.view(-1, self.flat_size)
        h = self.fc(h)
        h = F.relu(self.bn2(h))
        x = torch.matmul(h, obj_emb.t())
        return x


class GNNLinkPredict(nn.Module):
    def __init__(self):
        super(GNNLinkPredict, self).__init__()
        self.edge_set = None

    def forward(self, graph):
        raise NotImplementedError

    def get_edge_set(self, edge_index, edge_types):
        if self.edge_set is None:
            edge_list = torch.stack((edge_index[0], edge_index[1], edge_types))
            edge_list = edge_list.cpu().T.numpy().tolist()
            torch.cuda.empty_cache()
            self.edge_set = set([tuple(x) for x in edge_list])  # tuple(h, t, r)

    def _loss(self, head_embed, tail_embed, rel_embed, labels, scoring):
        score = scoring(head_embed, tail_embed, rel_embed)
        prediction_loss = F.binary_cross_entropy_with_logits(score, labels.float())
        return prediction_loss

    def _regularization(self, embs):
        loss = 0
        for emb in embs:
            loss += torch.mean(emb.pow(2))
        return loss


def sampling_edge_uniform(
    edge_index, edge_types, edge_set, sampling_rate, num_rels, label_smoothing=0.0, num_entities=1
):
    """
    Args:
        edge_index: edge index of graph
        edge_types:
        edge_set: set of all edges of the graph, (h, t, r)
        sampling_rate:
        num_rels:
        label_smoothing(Optional):
        num_entities (Optional):

    Returns:
        sampled_edges: sampled existing edges
        rels: types of smapled existing edges
        sampled_edges_all: existing edges with corrupted edges
        sampled_types_all: types of existing and corrupted edges
        labels: 0/1
    """
    num_edges = edge_index[0].shape[0]
    row, col = edge_index
    num_sampled_edges = int(num_edges * sampling_rate)

    selected_edges = np.random.choice(range(num_edges), num_sampled_edges, replace=False)
    row, col = row[selected_edges], col[selected_edges]
    sampled_edges = torch.stack([row, col])

    sampled_nodes = torch.unique(sampled_edges).cpu().numpy()

    heads = sampled_edges[0].cpu().numpy()
    tails = sampled_edges[1].cpu().numpy()
    rels = edge_types[selected_edges].cpu().numpy()
    torch.cuda.empty_cache()

    def to_set(head, tail, rel):
        triplets = np.stack((head, tail, rel), axis=0).T
        triplets_set = set([tuple(x) for x in triplets])
        return triplets_set

    corrupt_heads = np.random.choice(sampled_nodes, num_sampled_edges, replace=True)
    corrupt_heads = to_set(corrupt_heads, tails, rels)
    corrupt_tails = np.random.choice(sampled_nodes, num_sampled_edges, replace=True)
    corrupt_tails = to_set(heads, corrupt_tails, rels)
    corrupt_rels = np.random.choice(num_rels, num_sampled_edges, replace=True)
    corrupt_rels = to_set(heads, tails, corrupt_rels)

    corrupt_triplets = corrupt_heads.union(corrupt_tails).union(corrupt_rels)
    corrupt_triplets = corrupt_triplets.difference(edge_set)
    corrupt_triplets = torch.tensor(list(corrupt_triplets)).to(row.device).T

    _edge_index = corrupt_triplets[0:2]
    _edge_types = corrupt_triplets[2]

    sampled_edges_all = torch.cat((sampled_edges, _edge_index), dim=-1)
    edge_types_all = torch.cat((edge_types[selected_edges], _edge_types), dim=-1)
    labels = torch.tensor([1] * num_sampled_edges + [0] * _edge_index.shape[1]).to(row.device)
    if label_smoothing > 0:
        labels = (1.0 - label_smoothing) * labels + 1.0 / num_entities
    return sampled_edges, torch.from_numpy(rels), sampled_edges_all, edge_types_all, labels


def get_rank(scores, target):
    _, indices = torch.sort(scores, dim=1, descending=True)
    rank = (indices == target.view(-1, 1)).nonzero()[:, 1]
    return rank.view(-1)


def get_raw_rank(heads, tails, rels, embedding, rel_embedding, batch_size, scoring):
    test_size = heads.shape[0]
    num_batch = (test_size + batch_size - 1) // batch_size
    ranks = []
    for i in range(num_batch):
        start = batch_size * i
        end = start + batch_size
        scores = torch.sigmoid(scoring.predict(embedding[heads[start:end]], embedding, rel_embedding[rels[start:end]]))
        target = tails[start:end]
        rank = get_rank(scores, target).cpu().numpy()
        torch.cuda.empty_cache()
        ranks.append(rank)
    return np.concatenate(ranks).astype(np.float)


def get_filtered_rank(heads, tails, rels, embedding, rel_embedding, batch_size, seen_data):
    pass
